import unittest
from datetime import timedelta

from peewee import SqliteDatabase

from insomniac import db_models
from insomniac.actions_types import SourceType
from insomniac.db_models import get_ig_profile_by_profile_name, ProfileStatus, MODELS
from insomniac.utils import *

TEST_DATABASE_FILE = 'test.db'
test_db = SqliteDatabase(TEST_DATABASE_FILE, autoconnect=False)


class DatabaseTests(unittest.TestCase):

    def setUp(self):
        print("Creating test database with tables")
        db_models.db = test_db
        test_db.bind(MODELS)
        test_db.connect()
        test_db.create_tables(MODELS)
        test_db.close()

    def test_is_interacted(self):
        my_account1 = "my_account1"
        my_account2 = "my_account2"
        username1 = "username1"
        username2 = "username2"
        username3 = "username3"
        username4 = "username4"
        username5 = "username5"

        def job_interact(profile, session_id):
            print(COLOR_BOLD + f"Check users are not interacted by default" + COLOR_ENDC)
            assert profile.is_interacted(username1) is False

            print(COLOR_BOLD + f"Check users are not \"used to follow\" by default" + COLOR_ENDC)
            assert profile.used_to_follow(username1) is False

            print(COLOR_BOLD + f"Check interaction counted for {username1} after a \"get profile\"" + COLOR_ENDC)
            profile.log_get_profile_action(session_id, username1)
            assert profile.is_interacted(username1) is True

            print(COLOR_BOLD + f"Check interaction counted for {username2} after a like" + COLOR_ENDC)
            profile.log_like_action(session_id, username2, SourceType.BLOGGER.name, "some_blogger")
            assert profile.is_interacted(username2) is True

            print(COLOR_BOLD + f"Check interaction counted for {username3} after a comment" + COLOR_ENDC)
            profile.log_comment_action(session_id, username3, "Wow!", SourceType.BLOGGER.name, "some_blogger")
            assert profile.is_interacted(username3) is True

            print(COLOR_BOLD + f"Check interaction counted for {username4} after multiple actions" + COLOR_ENDC)
            profile.log_get_profile_action(session_id, username4)
            profile.log_like_action(session_id, username4, SourceType.BLOGGER.name, "some_blogger")
            profile.log_like_action(session_id, username4, SourceType.HASHTAG.name, "some_hashtag")
            profile.log_comment_action(session_id, username4, "Wow!", SourceType.PLACE.name, "some_place")
            assert profile.is_interacted(username4) is True

            print(COLOR_BOLD + f"Check interaction is NOT counted for {username5} after "
                               f"follow / story watch / unfollow / filter actions" + COLOR_ENDC)
            profile.log_follow_action(session_id, username5, None, None)
            profile.log_story_watch_action(session_id, username5, None, None)
            profile.log_unfollow_action(session_id, username5)
            profile.log_filter_action(session_id, username5)
            assert profile.is_interacted(username5) is False

            print(COLOR_BOLD + f"Check \"used to follow\" is True after following" + COLOR_ENDC)
            assert profile.used_to_follow(username5) is True
        self._run_inside_session(my_account1, job_interact)

        def job_interact(profile, _):
            print(COLOR_BOLD + f"Check interaction is not counted for {username1} under another account" + COLOR_ENDC)
            assert profile.is_interacted(username1) is False

            print(COLOR_BOLD + f"Check interaction is not counted for {username2} under another account" + COLOR_ENDC)
            assert profile.is_interacted(username2) is False

            print(COLOR_BOLD + f"Check interaction is not counted for {username3} under another account" + COLOR_ENDC)
            assert profile.is_interacted(username3) is False

            print(COLOR_BOLD + f"Check interaction is not counted for {username4} under another account" + COLOR_ENDC)
            assert profile.is_interacted(username4) is False

            print(COLOR_BOLD + f"Check interaction is not counted for {username5} under another account" + COLOR_ENDC)
            assert profile.is_interacted(username5) is False
        self._run_inside_session(my_account2, job_interact)

        def job_interact(profile, session_id):
            print(COLOR_BOLD + f"Check interaction is not counted for {username1} if was too long ago" + COLOR_ENDC)
            profile.log_like_action(session_id, username1, SourceType.BLOGGER.name, "some_blogger", datetime.now() - timedelta(hours=48))
            assert profile.is_interacted(username1) is True
            assert profile.is_interacted(username1, hours=24) is False
        self._run_inside_session(my_account2, job_interact)

    def test_is_filtered(self):
        my_account1 = "my_account1"
        my_account2 = "my_account2"
        username1 = "username1"
        username2 = "username2"

        def job_interact(profile, session_id):
            print(COLOR_BOLD + f"Check users are not filtered by default" + COLOR_ENDC)
            assert profile.is_filtered(username1) is False

            print(COLOR_BOLD + f"Check that filter works" + COLOR_ENDC)
            profile.log_filter_action(session_id, username1)
            assert profile.is_filtered(username1) is True

            print(COLOR_BOLD + f"Check that filter NOT works if filtered too long ago" + COLOR_ENDC)
            profile.log_filter_action(session_id, username2, datetime.now() - timedelta(hours=48))
            assert profile.is_filtered(username2) is True
            assert profile.is_filtered(username2, hours=24) is False
        self._run_inside_session(my_account1, job_interact)

        def job_interact(profile, _):
            print(COLOR_BOLD + f"Check that filter NOT works if filtered under another account" + COLOR_ENDC)
            assert profile.is_filtered(username1) is False
        self._run_inside_session(my_account2, job_interact)

    def test_following_status(self):
        my_account1 = "my_account1"
        my_account2 = "my_account2"
        username1 = "username1"

        def job_interact(profile, _):
            print(COLOR_BOLD + f"Check is_follow_me is None by default" + COLOR_ENDC)
            assert profile.is_follow_me(username1) is None

            print(COLOR_BOLD + f"Check do_i_follow is None by default" + COLOR_ENDC)
            assert profile.do_i_follow(username1) is None

            print(COLOR_BOLD + f"Check that last status is actual one" + COLOR_ENDC)
            profile.update_follow_status(username1, is_follow_me=True, do_i_follow_him=True)
            sleep(1)
            profile.update_follow_status(username1, is_follow_me=False, do_i_follow_him=True)
            assert profile.is_follow_me(username1) is False
            assert profile.do_i_follow(username1) is True
        self._run_inside_session(my_account1, job_interact)

        def job_interact(profile, _):
            print(COLOR_BOLD + f"Check that status is saved only for one account" + COLOR_ENDC)
            assert profile.is_follow_me(username1) is None
            assert profile.do_i_follow(username1) is None
        self._run_inside_session(my_account2, job_interact)

    def test_scraping(self):
        real_account_username = "real"
        scraper_account_username = "scraper"
        username1 = "username1"
        username2 = "username2"
        username3 = "username3"
        username4 = "username4"
        username5 = "username5"
        username6 = "username6"
        username7 = "username7"

        def job_scraper(profile, _):
            print(COLOR_BOLD + "Scraper: check accounts are not scraped by default" + COLOR_ENDC)
            assert profile.is_scrapped(username1, [real_account_username]) is False

            print("Scraper: publishing scrapped accounts")
            profile.publish_scrapped_account(username1, [real_account_username])
            profile.publish_scrapped_account(username1, [real_account_username])  # check with double insert
            profile.publish_scrapped_account(username2, [real_account_username])
            profile.publish_scrapped_account(username3, [real_account_username])
            profile.publish_scrapped_account(username4, [real_account_username])
            profile.publish_scrapped_account(username5, [real_account_username])
            profile.publish_scrapped_account(username6, [real_account_username])

            print(COLOR_BOLD + f"Scraper: check {username1} counts as scrapped" + COLOR_ENDC)
            assert profile.is_scrapped(username1, [real_account_username]) is True
            print(COLOR_BOLD + f"Scraper: check {username1} NOT counts as scrapped for another account" + COLOR_ENDC)
            assert profile.is_scrapped(username1, [real_account_username, "some_another_account"]) is False
        self._run_inside_session(scraper_account_username, job_scraper)

        def job_real(profile, session_id):
            print(COLOR_BOLD + "Real: check scraped accounts count is correct BEFORE interaction" + COLOR_ENDC)
            assert profile.count_scrapped_profiles_for_interaction() == 6

            print(COLOR_BOLD + "Real: check order is FIFO" + COLOR_ENDC)
            username = profile.get_scrapped_profile_for_interaction()
            assert username == username1

            print(COLOR_BOLD + "Real: check idempotence" + COLOR_ENDC)
            username = profile.get_scrapped_profile_for_interaction()
            assert username == username1

            print(COLOR_BOLD + "Real: check account is excluded after \"get profile action\"" + COLOR_ENDC)
            profile.log_get_profile_action(session_id, username1)
            username = profile.get_scrapped_profile_for_interaction()
            assert username == username2

            print(COLOR_BOLD + "Real: check account is excluded after like-interaction" + COLOR_ENDC)
            profile.log_like_action(session_id, username2, SourceType.BLOGGER.name, "some_blogger")
            username = profile.get_scrapped_profile_for_interaction()
            assert username == username3

            print(COLOR_BOLD + "Real: check account is excluded after follow-interaction" + COLOR_ENDC)
            profile.log_follow_action(session_id, username3, SourceType.HASHTAG.name, "some_hashtag")
            username = profile.get_scrapped_profile_for_interaction()
            assert username == username4

            print(COLOR_BOLD + "Real: check account is excluded after story-watch-interaction" + COLOR_ENDC)
            profile.log_comment_action(session_id, username4, "Wow!", SourceType.BLOGGER.name, "some_blogger")
            username = profile.get_scrapped_profile_for_interaction()
            assert username == username5

            print(COLOR_BOLD + "Real: check account is excluded after comment-interaction" + COLOR_ENDC)
            profile.log_comment_action(session_id, username5, "Wow!", SourceType.PLACE.name, "some_place")
            username = profile.get_scrapped_profile_for_interaction()
            assert username == username6

            print(COLOR_BOLD + "Real: check account is NOT excluded after unfollow / filter actions" + COLOR_ENDC)
            profile.log_unfollow_action(session_id, username4)
            profile.log_filter_action(session_id, username4)
            username = profile.get_scrapped_profile_for_interaction()
            assert username == username6

            profile.log_like_action(session_id, username6, SourceType.BLOGGER.name, "some_blogger")
            profile.log_like_action(session_id, username6, SourceType.BLOGGER.name, "some_blogger")  # double action check

            print(COLOR_BOLD + "Real: check scraped accounts count is correct AFTER interaction" + COLOR_ENDC)
            assert profile.count_scrapped_profiles_for_interaction() == 0
        self._run_inside_session(real_account_username, job_real)

        def job_scraper(profile, _):
            print("Scraper: publishing scrapped accounts")
            profile.publish_scrapped_account(username1, [real_account_username])
        self._run_inside_session(scraper_account_username, job_scraper)

        def job_real(profile, _):
            print(COLOR_BOLD + "Real: check account is still excluded after being interacted and scraped again" + COLOR_ENDC)
            username = profile.get_scrapped_profile_for_interaction()
            assert username is None
        self._run_inside_session(real_account_username, job_real)

        def job_scraper(profile, _):
            print("Scraper: publishing scrapped accounts")
            profile.publish_scrapped_account(username7, ["some_another_account"])
        self._run_inside_session(scraper_account_username, job_scraper)

        def job_real(profile, _):
            print(COLOR_BOLD + "Real: check account is excluded if scraped for another account" + COLOR_ENDC)
            username = profile.get_scrapped_profile_for_interaction()
            assert username is None
        self._run_inside_session(real_account_username, job_real)

        def job_scraper(profile, _):
            print("Scraper: publishing scrapped accounts")
            profile.publish_scrapped_account(username7, [real_account_username, "some_another_account"])
        self._run_inside_session(scraper_account_username, job_scraper)

        def job_real(profile, _):
            print(COLOR_BOLD + "Real: check account is accepted if scraped again but for real account" + COLOR_ENDC)
            username = profile.get_scrapped_profile_for_interaction()
            assert username == username7
        self._run_inside_session(real_account_username, job_real)

    def tearDown(self):
        print("Deleting test database")
        os.remove(TEST_DATABASE_FILE)

    def _run_inside_session(self, username, action):
        print(f"Starting session for {username}")
        profile = get_ig_profile_by_profile_name(username)
        session_id = profile.start_session(None, "", "", ProfileStatus.VALID, 2200, 500)
        print(f"session_id = {session_id}")
        action(profile, session_id)
        print(f"Ending session for {username}")
        profile.end_session(session_id)




from insomniac.action_runners import *
from insomniac.actions_types import TargetType
from insomniac.safely_runner import run_safely
from insomniac.utils import *


class InteractBySourceActionRunner(CoreActionsRunner):
    ACTION_ID = "interact"
    ACTION_ARGS = {
        "likes_count": {
            "help": "number of likes for each interacted user, 2 by default. "
                    "It can be a number (e.g. 2) or a range (e.g. 2-4)",
            'metavar': '2-4',
            "default": '2'
        },
        "likes_percentage": {
            "help": "likes given percentage of interacted users, 100 by default",
            "metavar": '50',
            "default": '100'
        },
        "follow_percentage": {
            "help": "follow given percentage of interacted users, 0 by default",
            "metavar": '50',
            "default": '0'
        },
        "interact": {
            "nargs": '+',
            "help": 'list of hashtags, usernames, or places. Usernames should start with \"@\" symbol. Places should '
                    'start with \"P-\" symbols. You can specify the way of interaction after a \"-\" sign: '
                    '@natgeo-followers, @natgeo-following, amazingtrips-top-likers, amazingtrips-recent-likers, '
                    'P-Paris-top-likers, P-Paris-recent-likers',
            "default": [],
            "metavar": ('amazingtrips-top-likers', '@natgeo-followers')
        },
        "interaction_users_amount": {
            "help": 'add this argument to select an amount of sources from the interact-list '
                    '(sources are randomized). It can be a number (e.g. 4) or a range (e.g. 3-8)',
            'metavar': '3-8'
        },
        "stories_count": {
            "help": 'number of stories to watch for each user, disabled by default. '
                    'It can be a number (e.g. 2) or a range (e.g. 2-4)',
            'metavar': '3-8'
        },
        "comment_percentage": {
            "help": "comment given percentage of interacted users, 0 by default",
            "metavar": '50',
            "default": '0'
        },
        "comments_list": {
            "nargs": '+',
            "help": 'list of comments you wish to comment on posts during interaction',
            "default": [],
            "metavar": ('WOW!', 'What a picture!')
        },
        "blacklist_profiles": {
            "nargs": '+',
            "help": 'list of profiles you wish to not interact with / scrape',
        }
    }

    likes_count = '2'
    follow_percentage = 0
    like_percentage = 100
    interact = []
    stories_count = '0'
    comment_percentage = 0
    comments_list = []

    def is_action_selected(self, args):
        return args.interact is not None and len(args.interact) > 0

    def reset_params(self):
        self.likes_count = '2'
        self.follow_percentage = 0
        self.like_percentage = 100
        self.interact = []
        self.stories_count = '0'
        self.comment_percentage = 0
        self.comments_list = []

    def set_params(self, args):
        self.reset_params()

        if args.likes_count is not None:
            self.likes_count = args.likes_count

        if args.stories_count is not None:
            self.stories_count = args.stories_count

        if args.interact is not None:
            self.interact = args.interact.copy()

            def is_source_hashtag(source):
                return source[0] != '@' and not source.startswith("P-")

            self.interact = [source if not is_source_hashtag(source) else ('#' + source) for source in self.interact]

        if args.follow_percentage is not None:
            self.follow_percentage = int(args.follow_percentage)

        if args.likes_percentage is not None:
            self.like_percentage = int(args.likes_percentage)

        if args.comment_percentage is not None:
            self.comment_percentage = int(args.comment_percentage)

        if args.comments_list is not None:
            self.comments_list = args.comments_list

        if args.interaction_users_amount is not None:
            if len(self.interact) > 0:
                users_amount = get_value(args.interaction_users_amount, "Interaction user amount {}", 100)

                if users_amount >= len(self.interact):
                    print("interaction-users-amount parameter is equal or higher then the users-interact list. "
                          "Choosing all list for interaction.")
                else:
                    amount_to_remove = len(self.interact) - users_amount
                    for i in range(0, amount_to_remove):
                        self.interact.remove(random.choice(self.interact))

    def run(self, device_wrapper, storage, session_state, on_action, is_limit_reached, is_passed_filters=None):
        from insomniac.action_runners.interact.action_handle_blogger import handle_blogger, extract_blogger_instructions
        from insomniac.action_runners.interact.action_handle_hashtag import handle_hashtag, extract_hashtag_instructions
        from insomniac.action_runners.interact.action_handle_place import handle_place, extract_place_instructions

        random.shuffle(self.interact)

        for source in self.interact:
            self.action_status = ActionStatus(ActionState.PRE_RUN)

            if source[0] == '@':
                is_myself = source[1:] == session_state.my_username
                print_timeless("")
                print(COLOR_BOLD + "Handle " + source + (is_myself and " (it\'s you)" or "") + COLOR_ENDC)
            elif source[0] == '#':
                print_timeless("")
                print(COLOR_BOLD + "Handle " + source + COLOR_ENDC)
            elif source.startswith("P-"):
                print_timeless("")
                print(COLOR_BOLD + "Handle " + source + COLOR_ENDC)

            @run_safely(device_wrapper=device_wrapper)
            def job():
                self.action_status.set(ActionState.RUNNING)
                if source[0] == '@':
                    source_name, instructions = extract_blogger_instructions(source)
                    handle_blogger(device_wrapper.get(),
                                   source_name[1:],  # drop "@"
                                   instructions,
                                   session_state,
                                   self.likes_count,
                                   self.stories_count,
                                   self.follow_percentage,
                                   self.like_percentage,
                                   self.comment_percentage,
                                   self.comments_list,
                                   storage,
                                   on_action,
                                   is_limit_reached,
                                   is_passed_filters,
                                   self.action_status)
                elif source[0] == '#':
                    source_name, instructions = extract_hashtag_instructions(source)
                    handle_hashtag(device_wrapper.get(),
                                   source_name[1:],  # drop "#"
                                   instructions,
                                   session_state,
                                   self.likes_count,
                                   self.stories_count,
                                   self.follow_percentage,
                                   self.like_percentage,
                                   self.comment_percentage,
                                   self.comments_list,
                                   storage,
                                   on_action,
                                   is_limit_reached,
                                   is_passed_filters,
                                   self.action_status)
                elif source.startswith("P-"):
                    source_name, instructions = extract_place_instructions(source[2:])
                    handle_place(device_wrapper.get(),
                                 source_name,
                                 instructions,
                                 session_state,
                                 self.likes_count,
                                 self.stories_count,
                                 self.follow_percentage,
                                 self.like_percentage,
                                 self.comment_percentage,
                                 self.comments_list,
                                 storage,
                                 on_action,
                                 is_limit_reached,
                                 is_passed_filters,
                                 self.action_status)
                self.action_status.set(ActionState.DONE)

            while not self.action_status.get() == ActionState.DONE:
                job()
                if self.action_status.get_limit() == ActionState.SOURCE_LIMIT_REACHED or \
                   self.action_status.get_limit() == ActionState.SESSION_LIMIT_REACHED:
                    break

            if self.action_status.get_limit() == ActionState.SOURCE_LIMIT_REACHED:
                continue

            if self.action_status.get_limit() == ActionState.SESSION_LIMIT_REACHED:
                break


class InteractByTargetsActionRunner(CoreActionsRunner):
    ACTION_ID = "interact_targets"
    ACTION_ARGS = {
        "interact_targets": {
            "help": "use this argument in order to interact with profiles from targets.txt",
            'metavar': 'True / False'
        },
        "targets_list": {
            "nargs": '+',
            "help": 'list of target-profiles you wish to interact with (in case you want to use a parameter '
                    'and not targets.txt file)',
            "default": [],
            "metavar": ('profile-A', 'profile-B')
        },
        "posts_urls_list": {
            "nargs": '+',
            "help": 'list of target-posts you wish to interact with (in case you want to use a parameter '
                    'and not targets.json file)',
            "default": [],
            "metavar": ('https://www.instagram.com/p/ID-a', 'https://www.instagram.com/p/ID-b')
        },
        "likes_count": {
            "help": "number of likes for each interacted user, 2 by default. "
                    "It can be a number (e.g. 2) or a range (e.g. 2-4)",
            'metavar': '2-4',
            "default": '2'
        },
        "follow_percentage": {
            "help": "follow given percentage of interacted users, 0 by default",
            "metavar": '50',
            "default": '0'
        },
        "likes_percentage": {
            "help": "likes given percentage of interacted users, 100 by default",
            "metavar": '50',
            "default": '100'
        },
        "stories_count": {
            "help": 'number of stories to watch for each user, disabled by default. '
                    'It can be a number (e.g. 2) or a range (e.g. 2-4)',
            'metavar': '3-8'
        },
        "comment_percentage": {
            "help": "comment given percentage of interacted users, 0 by default",
            "metavar": '50',
            "default": '0'
        },
        "comments_list": {
            "nargs": '+',
            "help": 'list of comments you wish to comment on posts during interaction',
            "default": [],
            "metavar": ('WOW!', 'What a picture!')
        },
        "blacklist_profiles": {
            "nargs": '+',
            "help": 'list of profiles you wish to not interact with / scrape',
        }
    }

    likes_count = '2'
    follow_percentage = 0
    like_percentage = 100
    stories_count = '0'
    comment_percentage = 0
    comments_list = []

    def is_action_selected(self, args):
        return args.interact_targets is not None

    def reset_params(self):
        self.likes_count = '2'
        self.follow_percentage = 0
        self.like_percentage = 100
        self.stories_count = '0'
        self.comment_percentage = 0
        self.comments_list = []

    def set_params(self, args):
        self.reset_params()

        if args.likes_count is not None:
            self.likes_count = args.likes_count

        if args.stories_count is not None:
            self.stories_count = args.stories_count

        if args.follow_percentage is not None:
            self.follow_percentage = int(args.follow_percentage)

        if args.likes_percentage is not None:
            self.like_percentage = int(args.likes_percentage)

        if args.comment_percentage is not None:
            self.comment_percentage = int(args.comment_percentage)

        if args.comments_list is not None:
            self.comments_list = args.comments_list

    def run(self, device_wrapper, storage, session_state, on_action, is_limit_reached, is_passed_filters=None):
        from insomniac.action_runners.interact.action_handle_target import handle_target

        target, target_type = storage.get_target()
        while target is not None:
            self.action_status = ActionStatus(ActionState.PRE_RUN)

            print_timeless("")
            print(COLOR_BOLD + f"Handle {'@' if target_type == TargetType.USERNAME else ''}" + target + COLOR_ENDC)

            @run_safely(device_wrapper=device_wrapper)
            def job():
                self.action_status.set(ActionState.RUNNING)
                handle_target(device_wrapper.get(),
                              target,
                              target_type,
                              session_state,
                              self.likes_count,
                              self.stories_count,
                              self.follow_percentage,
                              self.like_percentage,
                              self.comment_percentage,
                              self.comments_list,
                              storage,
                              on_action,
                              is_limit_reached,
                              is_passed_filters,
                              self.action_status)

                self.action_status.set(ActionState.DONE)

            while not self.action_status.get() == ActionState.DONE:
                job()
                if self.action_status.get_limit() == ActionState.SOURCE_LIMIT_REACHED or \
                        self.action_status.get_limit() == ActionState.SESSION_LIMIT_REACHED:
                    break

            if self.action_status.get_limit() == ActionState.SOURCE_LIMIT_REACHED:
                continue

            if self.action_status.get_limit() == ActionState.SESSION_LIMIT_REACHED:
                break

            target, target_type = storage.get_target()

        if target is None:
            print("There are no more new targets to interact with in the database (all been already interacted / filtered).")
            print("If you wish to continue interacting with targets, add new targets to the database using scrapping / targets.txt "
                  "or use the reinteract-after & refilter-after parameters in order to interact with the targets that are already loaded "
                  "in the database.")



from enum import Enum, unique
from os import listdir
from random import uniform
from re import search
from typing import Optional

from insomniac.sleeper import sleeper
from insomniac.utils import *

# How long we're waiting until UI element appears (loading content + animation)
UI_TIMEOUT_LONG = 5
UI_TIMEOUT_SHORT = 1

SCREEN_RECORDS_PATH = "screen_records"


def create_device(is_old, device_id, app_id, typewriter):
    print("Using uiautomator v" + ("1" if is_old else "2"))
    try:
        return DeviceFacade(is_old, device_id, app_id, typewriter)
    except ImportError as e:
        print(COLOR_FAIL + str(e) + COLOR_ENDC)
        return None


class DeviceFacade:
    deviceV1 = None  # uiautomator
    deviceV2 = None  # uiautomator2
    width = None
    height = None
    device_id = None
    app_id = None
    typewriter = None

    def __init__(self, is_old, device_id, app_id, typewriter):
        self.device_id = device_id
        self.app_id = app_id
        self.typewriter = typewriter
        if is_old:
            try:
                import uiautomator
                self.deviceV1 = uiautomator.device if device_id is None else uiautomator.Device(device_id)
            except ImportError:
                raise ImportError("Please install uiautomator: pip3 install uiautomator")
        else:
            try:
                import uiautomator2
                self.deviceV2 = uiautomator2.connect() if device_id is None else uiautomator2.connect(device_id)
            except ImportError:
                raise ImportError("Please install uiautomator2: pip3 install uiautomator2")

    def is_old(self):
        return self.deviceV1 is not None

    def find(self, *args, **kwargs):
        if self.deviceV1 is not None:
            import uiautomator
            try:
                view = self.deviceV1(*args, **kwargs)
            except uiautomator.JsonRPCError as e:
                raise DeviceFacade.JsonRpcError(e)
            return DeviceFacade.View(is_old=True, view=view, device=self)
        else:
            import uiautomator2
            try:
                view = self.deviceV2(*args, **kwargs)
            except uiautomator2.JSONRPCError as e:
                raise DeviceFacade.JsonRpcError(e)
            return DeviceFacade.View(is_old=False, view=view, device=self)

    def back(self):
        """
        Press back and check that UI hierarchy was changed. If it didn't change, it means that back press didn't work.
        So, we try to press back several times until it is finally changed.
        """
        max_attempts = 5

        def dump_hierarchy():
            if self.deviceV1 is not None:
                return self.deviceV1.dump()
            else:
                return self.deviceV2.dump_hierarchy()

        def normalize(hierarchy):
            """
            Remove all texts from hierarchy. It may contain some changing data, e.g. current time.
            """
            return re.sub(r'text=".*"', 'text=""', hierarchy)

        succeed = False
        attempts = 0
        while not succeed:
            if attempts >= max_attempts:
                print(COLOR_FAIL + f"Tried to press back {attempts} times with no success. Will proceed next..." +
                      COLOR_ENDC)
                break
            hierarchy_before = normalize(dump_hierarchy())
            self._press_back()
            hierarchy_after = normalize(dump_hierarchy())
            succeed = hierarchy_before != hierarchy_after
            if not succeed:
                print(COLOR_OKGREEN + "Pressed back but nothing changed on the screen. Will try again." + COLOR_ENDC)
                sleeper.random_sleep()
            attempts += 1

    def _press_back(self):
        if self.deviceV1 is not None:
            self.deviceV1.press.back()
        else:
            self.deviceV2.press("back")

    def open_notifications(self):
        os.popen("adb" + ("" if self.device_id is None else " -s " + self.device_id) +
                 f" shell service call statusbar 1 {self.app_id}").close()

    def hide_notifications(self):
        os.popen("adb" + ("" if self.device_id is None else " -s " + self.device_id) +
                 f" shell service call statusbar 2 {self.app_id}").close()

    def screen_click(self, place):
        w, h = self._get_screen_size()
        if place == DeviceFacade.Place.RIGHT:
            left = int(w * 3 / 4)
            top = int(h / 2)
        else:
            return

        self.screen_click_by_coordinates(left, top)

    def screen_click_by_coordinates(self, left, top):
        if self.deviceV1 is not None:
            self.deviceV1.click(left, top)
        else:
            self.deviceV2.click(left, top)

    def screenshot(self, path):
        if self.deviceV1 is not None:
            self.deviceV1.screenshot(path)
        else:
            self.deviceV2.screenshot(path)

    def start_screen_record(self, fps=10):
        """Available for uiautomator2 only"""
        if self.deviceV1 is not None:
            print(COLOR_FAIL + "Screen record doesn't work when you use the --old flag" + COLOR_ENDC)
        else:
            if not os.path.exists(SCREEN_RECORDS_PATH):
                os.makedirs(SCREEN_RECORDS_PATH)
            mp4_files = [f for f in listdir(SCREEN_RECORDS_PATH) if f.endswith(".mp4")]
            if mp4_files:
                last_mp4 = mp4_files[-1]
                debug_number = "{0:0=4d}".format(int(last_mp4[-8:-4]) + 1)
            else:
                debug_number = "0000"
            output = os.path.join(SCREEN_RECORDS_PATH, f"debug_{debug_number}.mp4")
            try:
                self.deviceV2.screenrecord(output, fps)
            except ModuleNotFoundError:
                print(COLOR_FAIL + "To use screen recording please install additional packages:" + COLOR_ENDC)
                print(COLOR_FAIL + COLOR_BOLD +
                      'pip3 install -U "uiautomator2[image]" -i https://pypi.doubanio.com/simple' + COLOR_ENDC)
                return
            print(COLOR_OKGREEN + f'Started screen recording: it will be saved as "{output}".' + COLOR_ENDC)

    def stop_screen_record(self):
        """Available for uiautomator2 only"""
        if self.deviceV1 is not None:
            return

        try:
            is_recorded = self.deviceV2.screenrecord.stop()
        except ModuleNotFoundError:
            is_recorded = False
        if is_recorded:
            if not os.path.exists(SCREEN_RECORDS_PATH):
                return
            mp4_files = [f for f in listdir(SCREEN_RECORDS_PATH) if f.endswith(".mp4")]
            if mp4_files:
                last_mp4 = mp4_files[-1]
                path = os.path.join(SCREEN_RECORDS_PATH, last_mp4)
                print(COLOR_OKGREEN + f'Screen recorder has been stopped successfully! Saved as "{path}".' + COLOR_ENDC)

    def dump_hierarchy(self, path):
        if self.deviceV1 is not None:
            xml_dump = self.deviceV1.dump()
        else:
            xml_dump = self.deviceV2.dump_hierarchy()

        with open(path, 'w', encoding="utf-8") as outfile:
            outfile.write(xml_dump)

    def is_screen_on(self):
        return self.get_info()["screenOn"]

    def press_power(self):
        if self.deviceV1 is not None:
            self.deviceV1.press.power()
        else:
            self.deviceV2.press("power")

    def is_screen_locked(self):
        cmd = f"adb {'' if self.device_id is None else ('-s '+ self.device_id)} shell dumpsys window"

        cmd_res = subprocess.run(cmd, stdout=PIPE, stderr=PIPE, shell=True, encoding="utf8")
        data = cmd_res.stdout.strip()
        flag = search("mDreamingLockscreen=(true|false)", data)
        return True if flag.group(1) == "true" else False

    def is_alive(self):
        if self.deviceV1 is not None:
            return self.deviceV1.alive()
        else:
            return self.deviceV2._is_alive()

    def wake_up(self):
        """ Make sure agent is alive or bring it back up before starting. """
        attempts = 0
        while not self.is_alive() and attempts < 5:
            self.get_info()
            attempts += 1

    def unlock(self):
        self.swipe(DeviceFacade.Direction.TOP, 0.8)
        if self.is_screen_locked():
            self.swipe(DeviceFacade.Direction.RIGHT, 0.8)

    def screen_off(self):
        if self.deviceV1 is not None:
            self.deviceV1.screen.off()
        else:
            self.deviceV2.screen_off()

    def swipe(self, direction: "DeviceFacade.Direction", scale=0.5):
        """Swipe finger in the `direction`.
        Scale is the sliding distance. Default to 50% of the screen width
        """

        if self.deviceV1 is not None:
            def _swipe(_from, _to):
                self.deviceV1.swipe(_from[0], _from[1], _to[0], _to[1])

            lx, ly = 0, 0
            rx, ry = self._get_screen_size()

            width, height = rx - lx, ry - ly

            h_offset = int(width * (1 - scale)) // 2
            v_offset = int(height * (1 - scale)) // 2

            left = lx + h_offset, ly + height // 2
            up = lx + width // 2, ly + v_offset
            right = rx - h_offset, ly + height // 2
            bottom = lx + width // 2, ry - v_offset

            if direction == DeviceFacade.Direction.TOP:
                _swipe(bottom, up)
            elif direction == DeviceFacade.Direction.RIGHT:
                _swipe(left, right)
            elif direction == DeviceFacade.Direction.LEFT:
                _swipe(right, left)
            elif direction == DeviceFacade.Direction.BOTTOM:
                _swipe(up, bottom)
        else:
            swipe_dir = ""
            if direction == DeviceFacade.Direction.TOP:
                swipe_dir = "up"
            elif direction == DeviceFacade.Direction.RIGHT:
                swipe_dir = "right"
            elif direction == DeviceFacade.Direction.LEFT:
                swipe_dir = "left"
            elif direction == DeviceFacade.Direction.BOTTOM:
                swipe_dir = "down"
            self.deviceV2.swipe_ext(swipe_dir, scale=scale)

    def swipe_points(self, sx, sy, ex, ey):
        if self.deviceV1 is not None:
            import uiautomator
            try:
                self.deviceV1.swipePoints([[sx, sy], [ex, ey]])
            except uiautomator.JsonRPCError as e:
                raise DeviceFacade.JsonRpcError(e)
        else:
            import uiautomator2
            try:
                self.deviceV2.swipe_points([[sx, sy], [ex, ey]], uniform(0.2, 0.6))
            except uiautomator2.JSONRPCError as e:
                raise DeviceFacade.JsonRpcError(e)

    def get_info(self):
        if self.deviceV1 is not None:
            return self.deviceV1.info
        else:
            return self.deviceV2.info

    def is_keyboard_open(self):
        cmd = f"adb {'' if self.device_id is None else ('-s '+ self.device_id)} shell dumpsys input_method"

        cmd_res = subprocess.run(cmd, stdout=PIPE, stderr=PIPE, shell=True, encoding="utf8")
        data = cmd_res.stdout.strip()
        flag = search("mInputShown=(true|false)", data)
        if flag is not None:
            return True if flag.group(1) == "true" else False
        return False

    def close_keyboard(self):
        print("Closing keyboard...")
        if self.is_keyboard_open():
            print("Keyboard is open, closing it by pressing back")
            self.back()
            print("Verifying again that keyboard is closed")
            if self.is_keyboard_open():
                print(COLOR_FAIL + "Keyboard is open and couldn't be closed for some reason" + COLOR_ENDC)
            else:
                print("The device keyboard is closed now.")

            return

        print("The device keyboard is already closed.")

    def _get_screen_size(self):
        if self.width is not None and self.height is not None:
            return self.width, self.height

        if self.deviceV1 is not None:
            self.width = self.deviceV1.info['displayWidth']
            self.height = self.deviceV1.info['displayHeight']
        else:
            self.width = self.deviceV2.info['displayWidth']
            self.height = self.deviceV2.info['displayHeight']

        return self.width, self.height

    class View:
        device = None
        viewV1 = None  # uiautomator
        viewV2 = None  # uiautomator2

        def __init__(self, is_old, view, device):
            self.device = device
            if is_old:
                self.viewV1 = view
            else:
                self.viewV2 = view

        def __iter__(self):
            children = []
            if self.viewV1 is not None:
                import uiautomator
                try:
                    for item in self.viewV1:
                        children.append(DeviceFacade.View(is_old=True, view=item, device=self.device))
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    for item in self.viewV2:
                        children.append(DeviceFacade.View(is_old=False, view=item, device=self.device))
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            return iter(children)

        def child(self, *args, **kwargs):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    view = self.viewV1.child(*args, **kwargs)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=True, view=view, device=self.device)
            else:
                import uiautomator2
                try:
                    view = self.viewV2.child(*args, **kwargs)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=False, view=view, device=self.device)

        def right(self, *args, **kwargs) -> Optional['DeviceFacade.View']:
            if self.viewV1 is not None:
                import uiautomator
                try:
                    view = self.viewV1.right(*args, **kwargs)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=True, view=view, device=self.device)
            else:
                import uiautomator2
                try:
                    view = self.viewV2.right(*args, **kwargs)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=False, view=view, device=self.device)

        def left(self, *args, **kwargs):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    view = self.viewV1.left(*args, **kwargs)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=True, view=view, device=self.device)
            else:
                import uiautomator2
                try:
                    view = self.viewV2.left(*args, **kwargs)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=False, view=view, device=self.device)

        def up(self, *args, **kwargs):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    view = self.viewV1.up(*args, **kwargs)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=True, view=view, device=self.device)
            else:
                import uiautomator2
                try:
                    view = self.viewV2.up(*args, **kwargs)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=False, view=view, device=self.device)

        def down(self, *args, **kwargs):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    view = self.viewV1.down(*args, **kwargs)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=True, view=view, device=self.device)
            else:
                import uiautomator2
                try:
                    view = self.viewV2.down(*args, **kwargs)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return DeviceFacade.View(is_old=False, view=view, device=self.device)

        def click(self, mode=None, ignore_if_missing=False):
            if ignore_if_missing and not self.exists(quick=True):
                return

            mode = DeviceFacade.Place.WHOLE if mode is None else mode
            if mode == DeviceFacade.Place.WHOLE:
                x_offset = uniform(0.15, 0.85)
                y_offset = uniform(0.15, 0.85)

            elif mode == DeviceFacade.Place.LEFT:
                x_offset = uniform(0.15, 0.4)
                y_offset = uniform(0.15, 0.85)

            elif mode == DeviceFacade.Place.CENTER:
                x_offset = uniform(0.4, 0.6)
                y_offset = uniform(0.15, 0.85)

            elif mode == DeviceFacade.Place.RIGHT:
                x_offset = uniform(0.6, 0.85)
                y_offset = uniform(0.15, 0.85)

            else:
                x_offset = 0.5
                y_offset = 0.5

            if self.viewV1 is not None:
                import uiautomator
                try:
                    self.viewV1.click.wait()
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    self.viewV2.click(UI_TIMEOUT_LONG, offset=(x_offset, y_offset))
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def long_click(self):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    self.viewV1.long_click()
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    self.viewV2.long_click()
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def double_click(self, padding=0.3):
            """
            Double click randomly in the selected view using padding
            padding: % of how far from the borders we want the double click to happen.
            """

            if self.viewV1 is not None:
                self._double_click_v1()
            else:
                self._double_click_v2(padding)

        def scroll(self, direction):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    if direction == DeviceFacade.Direction.TOP:
                        self.viewV1.scroll.toBeginning(max_swipes=1)
                    else:
                        self.viewV1.scroll.toEnd(max_swipes=1)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    if direction == DeviceFacade.Direction.TOP:
                        self.viewV2.scroll.toBeginning(max_swipes=1)
                    else:
                        self.viewV2.scroll.toEnd(max_swipes=1)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def swipe(self, direction):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    if direction == DeviceFacade.Direction.TOP:
                        self.viewV1.fling.toBeginning(max_swipes=5)
                    else:
                        self.viewV1.fling.toEnd(max_swipes=5)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    if direction == DeviceFacade.Direction.TOP:
                        self.viewV2.fling.toBeginning(max_swipes=5)
                    else:
                        self.viewV2.fling.toEnd(max_swipes=5)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def exists(self, quick=False):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    return self.viewV1.exists
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    return self.viewV2.exists(UI_TIMEOUT_SHORT if quick else UI_TIMEOUT_LONG)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def wait(self):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    self.deviceV1.wait.idle()
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
                return True
            else:
                import uiautomator2
                try:
                    return self.viewV2.wait(timeout=UI_TIMEOUT_LONG)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def get_bounds(self):
            if self.viewV1 is not None:
                import uiautomator
                try:
                    return self.viewV1.bounds
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    return self.viewV2.info['bounds']
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def get_text(self, retry=True):
            max_attempts = 1 if not retry else 3
            attempts = 0

            if self.viewV1 is not None:
                import uiautomator
                while attempts < max_attempts:
                    attempts += 1
                    try:
                        text = self.viewV1.text
                        if text is None:
                            if attempts < max_attempts:
                                print(COLOR_REPORT + "Could not get text. Waiting 2 seconds and trying again..." + COLOR_ENDC)
                                sleep(2)  # wait 2 seconds and retry
                                continue
                        else:
                            return text
                    except uiautomator.JsonRPCError as e:
                        raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                while attempts < max_attempts:
                    attempts += 1
                    try:
                        text = self.viewV2.info['text']
                        if text is None:
                            if attempts < max_attempts:
                                print(COLOR_REPORT + "Could not get text. "
                                                     "Waiting 2 seconds and trying again..." + COLOR_ENDC)
                                sleep(2)  # wait 2 seconds and retry
                                continue
                        else:
                            return text
                    except uiautomator2.JSONRPCError as e:
                        raise DeviceFacade.JsonRpcError(e)

            print(COLOR_FAIL + f"Attempted to get text {attempts} times. You may have a slow network or are "
                               f"experiencing another problem." + COLOR_ENDC)
            return ""

        def get_selected(self) -> bool:
            if self.viewV1 is not None:
                import uiautomator
                try:
                    self.viewV1.info["selected"]
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    return self.viewV2.info["selected"]
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def is_enabled(self) -> bool:
            if self.viewV1 is not None:
                import uiautomator
                try:
                    self.viewV1.info["enabled"]
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    return self.viewV2.info["enabled"]
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def set_text(self, text):
            if self.device.typewriter.write(self, text):
                return
            if self.viewV1 is not None:
                import uiautomator
                try:
                    self.viewV1.set_text(text)
                except uiautomator.JsonRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)
            else:
                import uiautomator2
                try:
                    self.viewV2.set_text(text)
                except uiautomator2.JSONRPCError as e:
                    raise DeviceFacade.JsonRpcError(e)

        def _double_click_v1(self):
            import uiautomator
            config = self.device.deviceV1.server.jsonrpc.getConfigurator()
            config['actionAcknowledgmentTimeout'] = 40
            self.device.deviceV1.server.jsonrpc.setConfigurator(config)
            try:
                self.viewV1.click()
                self.viewV1.click()
            except uiautomator.JsonRPCError as e:
                raise DeviceFacade.JsonRpcError(e)
            config['actionAcknowledgmentTimeout'] = 3000
            self.device.deviceV1.server.jsonrpc.setConfigurator(config)

        def _double_click_v2(self, padding):
            import uiautomator2
            visible_bounds = self.get_bounds()
            horizontal_len = visible_bounds["right"] - visible_bounds["left"]
            vertical_len = visible_bounds["bottom"] - visible_bounds["top"]
            horizintal_padding = int(padding * horizontal_len)
            vertical_padding = int(padding * vertical_len)
            random_x = int(
                uniform(
                    visible_bounds["left"] + horizintal_padding,
                    visible_bounds["right"] - horizintal_padding,
                )
            )
            random_y = int(
                uniform(
                    visible_bounds["top"] + vertical_padding,
                    visible_bounds["bottom"] - vertical_padding,
                )
            )
            time_between_clicks = uniform(0.050, 0.200)
            try:
                self.device.deviceV2.double_click(random_x, random_y, duration=time_between_clicks)
            except uiautomator2.JSONRPCError as e:
                raise DeviceFacade.JsonRpcError(e)

    @unique
    class Direction(Enum):
        TOP = 0
        BOTTOM = 1
        RIGHT = 2
        LEFT = 3

    @unique
    class Place(Enum):
        # TODO: add more places
        RIGHT = 0
        WHOLE = 1
        CENTER = 2
        BOTTOM = 3
        LEFT = 4

    class JsonRpcError(Exception):
        pass



from insomniac import db_models
from insomniac.actions_types import TargetType
from insomniac.database_engine import *
from insomniac.db_models import get_ig_profile_by_profile_name, ProfileStatus
from insomniac.utils import *

FILENAME_WHITELIST = "whitelist.txt"
FILENAME_BLACKLIST = "blacklist.txt"
FILENAME_TARGETS = "targets.txt"


STORAGE_ARGS = {
    "reinteract_after": {
        "help": "set a time (in hours) to wait before re-interact with an already interacted profile, "
                "disabled by default (won't interact again). "
                "It can be a number (e.g. 48) or a range (e.g. 50-80)",
        "metavar": "150"
    },
    "refilter_after": {
        "help": "set a time (in hours) to wait before re-filter an already filtered profile, "
                "disabled by default (will drop the profile and won't filter again). "
                "It can be a number (e.g. 48) or a range (e.g. 50-80)",
        "metavar": "150"
    },
    "recheck_follow_status_after": {
        "help": "set a time (in hours) to wait before re-check follow status of a profile, "
                "disabled by default (will check every time when needed)."
                "It can be a number (e.g. 48) or a range (e.g. 50-80)",
        "metavar": "150"
    }
}


IS_USING_DATABASE = False


def database_api(func):
    def wrap(*args, **kwargs):
        if IS_USING_DATABASE:
            return func(*args, **kwargs)

        return None
    return wrap


class Storage:
    profile = None
    scrape_for_account_list = []
    recheck_follow_status_after = None
    whitelist = []
    blacklist = []
    reinteract_after = None
    refilter_after = None
    profiles_targets_list_from_parameters = []
    url_targets_list_from_parameters = []

    def _reset_state(self):
        global IS_USING_DATABASE
        IS_USING_DATABASE = False
        self.profile = None
        self.scrape_for_account_list = []
        self.recheck_follow_status_after = None
        self.whitelist = []
        self.blacklist = []
        self.reinteract_after = None
        self.refilter_after = None
        self.profiles_targets_list_from_parameters = []
        self.url_targets_list_from_parameters = []

    def __init__(self, my_username, args):
        db_models.init()

        self._reset_state()

        if my_username is None:
            print(COLOR_FAIL + "No username, so the script won't read/write from the database" + COLOR_ENDC)
            return

        global IS_USING_DATABASE
        IS_USING_DATABASE = True

        self.profile = get_ig_profile_by_profile_name(my_username)
        scrape_for_account = args.__dict__.get('scrape_for_account', [])
        self.scrape_for_account_list = scrape_for_account if isinstance(scrape_for_account, list) else [scrape_for_account]
        if args.reinteract_after is not None:
            self.reinteract_after = get_value(args.reinteract_after, "Re-interact after {} hours", 168)
        if args.refilter_after is not None:
            self.refilter_after = get_value(args.refilter_after, "Re-filter after {} hours", 168)
        if args.recheck_follow_status_after is not None:
            self.recheck_follow_status_after = get_value(args.recheck_follow_status_after, "Re-check follow status after {} hours", 168)
        self.profiles_targets_list_from_parameters = args.__dict__.get('targets_list', [])
        self.url_targets_list_from_parameters = args.__dict__.get('posts_urls_list', [])
        whitelist_from_parameters = args.__dict__.get('whitelist_profiles', None)
        blacklist_from_parameters = args.__dict__.get('blacklist_profiles', None)

        # Whitelist and Blacklist
        try:
            with open(FILENAME_WHITELIST, encoding="utf-8") as file:
                self.whitelist = [line.rstrip() for line in file]
        except FileNotFoundError:
            print_debug("No whitelist provided")

        try:
            with open(FILENAME_BLACKLIST, encoding="utf-8") as file:
                self.blacklist = [line.rstrip() for line in file]
        except FileNotFoundError:
            print_debug("No blacklist provided")

        if whitelist_from_parameters is not None:
            if isinstance(whitelist_from_parameters, list) and len(whitelist_from_parameters) > 0:
                print("Loading whitelist from profiles_whitelist parameter...")
                self.whitelist.extend(whitelist_from_parameters)

        if blacklist_from_parameters is not None:
            if isinstance(blacklist_from_parameters, list) and len(blacklist_from_parameters) > 0:
                print("Loading blacklist from profiles_blacklist parameter...")
                self.blacklist.extend(blacklist_from_parameters)

        # Print meta data
        if len(self.profiles_targets_list_from_parameters) > 0 or len(self.url_targets_list_from_parameters) > 0:
            count = len(self.profiles_targets_list_from_parameters) + len(self.url_targets_list_from_parameters)
            print(f"Profiles and posts to interact from args: {count}")
        count_from_file = self._count_targets_from_file()
        if count_from_file > 0:
            print(f"Profiles and posts to interact from targets file: {count_from_file}")
        count_from_scrapping = self.profile.count_scrapped_profiles_for_interaction()
        if count_from_scrapping > 0:
            print(f"Profiles to interact from scrapping: {count_from_scrapping}")

    @database_api
    def start_session(self, app_id, app_version, args, followers_count, following_count):
        session_id = self.profile.start_session(app_id, app_version, args, ProfileStatus.VALID,
                                                followers_count, following_count)
        return session_id

    @database_api
    def end_session(self, session_id):
        self.profile.end_session(session_id)

    @database_api
    def check_user_was_interacted(self, username):
        return self.profile.is_interacted(username, hours=self.reinteract_after)

    @database_api
    def check_user_was_interacted(self, username):
        return self.profile.is_interacted(username, hours=72)

    @database_api
    def check_user_was_scrapped(self, username):
        return self.profile.is_scrapped(username, self.scrape_for_account_list)

    @database_api
    def check_user_was_filtered(self, username):
        return self.profile.is_filtered(username, hours=self.refilter_after)

    @database_api
    def get_following_status(self, username):
        if not self.profile.used_to_follow(username):
            return FollowingStatus.NONE
        return FollowingStatus.FOLLOWED if self.profile.do_i_follow(username) else FollowingStatus.UNFOLLOWED

    @database_api
    def is_profile_follows_me_by_cache(self, username):
        """
        Return True if and only if "username" follows me and the last check was within
        "recheck_follow_status_after" hours.
        """
        if self.recheck_follow_status_after is None:
            return False
        return self.profile.is_follow_me(username, hours=self.recheck_follow_status_after) is True

    @database_api
    def update_follow_status(self, username, is_follow_me, do_i_follow_him):
        self.profile.update_follow_status(username, is_follow_me, do_i_follow_him)

    @database_api
    def log_get_profile_action(self, session_id, username):
        self.profile.log_get_profile_action(session_id, username)

    @database_api
    def log_like_action(self, session_id, username, source_type, source_name):
        self.profile.log_like_action(session_id, username, source_type, source_name)

    @database_api
    def log_follow_action(self, session_id, username, source_type, source_name):
        self.profile.log_follow_action(session_id, username, source_type, source_name)

    @database_api
    def log_story_watch_action(self, session_id, username, source_type, source_name):
        self.profile.log_story_watch_action(session_id, username, source_type, source_name)

    @database_api
    def log_comment_action(self, session_id, username, comment, source_type, source_name):
        self.profile.log_comment_action(session_id, username, comment, source_type, source_name)

    @database_api
    def log_direct_message_action(self, session_id, username, message):
        self.profile.log_direct_message_action(session_id, username, message)

    @database_api
    def log_unfollow_action(self, session_id, username):
        self.profile.log_unfollow_action(session_id, username)

    @database_api
    def log_scrape_action(self, session_id, username, source_type, source_name):
        self.profile.log_scrape_action(session_id, username, source_type, source_name)

    @database_api
    def log_filter_action(self, session_id, username):
        self.profile.log_filter_action(session_id, username)

    @database_api
    def log_change_profile_info_action(self, session_id, profile_pic_url, name, description):
        self.profile.log_change_profile_info_action(session_id, profile_pic_url, name, description)

    @database_api
    def publish_scrapped_account(self, username):
        self.profile.publish_scrapped_account(username, self.scrape_for_account_list)

    def get_target(self):
        """
        Get a target from args (users/posts) -> OR from targets file (users/posts) -> OR from scrapping (only users).
        Picks only not yet interacted targets.

        :returns: target and type
        """
        # From args
        try:
            return self.profiles_targets_list_from_parameters.pop(0), TargetType.USERNAME
        except IndexError:
            pass

        try:
            return self.url_targets_list_from_parameters.pop(0), TargetType.URL
        except IndexError:
            pass

        # From file
        try:
            with open(FILENAME_TARGETS, "r+", encoding="utf-8") as file:
                lines = [line.rstrip() for line in file]

                for i, line in enumerate(lines):
                    # Skip comments
                    if line.startswith("#"):
                        continue

                    # Skip already interacted
                    if "DONE" in line:
                        continue

                    data = line.strip()
                    if data.startswith("https://"):
                        target_type = TargetType.URL
                    else:
                        target_type = TargetType.USERNAME
                    lines[i] += " - DONE"
                    file.truncate(0)
                    file.seek(0)
                    file.write("\n".join(lines))
                    return data, target_type
        except FileNotFoundError:
            pass

        # From scrapping
        scrapped_profile = self.profile.get_scrapped_profile_for_interaction()
        if scrapped_profile is not None:
            return scrapped_profile, TargetType.USERNAME
        return None, None

    def is_user_in_whitelist(self, username):
        return username in self.whitelist

    def is_user_in_blacklist(self, username):
        return username in self.blacklist

    def _count_targets_from_file(self):
        count = 0
        try:
            with open(FILENAME_TARGETS, encoding="utf-8") as file:
                lines = [line.rstrip() for line in file]

                for i, line in enumerate(lines):
                    # Skip comments
                    if line.startswith("#"):
                        continue

                    # Skip already interacted
                    if "DONE" in line:
                        continue

                    count += 1
        except FileNotFoundError:
            pass
        return count


@unique
class FollowingStatus(Enum):
    NONE = 0
    FOLLOWED = 1
    UNFOLLOWED = 2



from collections import namedtuple
from enum import unique, Enum

GetProfileAction = namedtuple('GetProfileAction', 'user')
LikeAction = namedtuple('LikeAction', 'source_name source_type user')
FollowAction = namedtuple('FollowAction', 'source_name source_type user')
StoryWatchAction = namedtuple('StoryWatchAction', 'source_name source_type user')
CommentAction = namedtuple('CommentAction', 'source_name source_type user comment')
DirectMessageAction = namedtuple('DirectMessageAction', 'user message')
UnfollowAction = namedtuple('UnfollowAction', 'user')
ScrapeAction = namedtuple('ScrapeAction', 'source_name source_type user')
FilterAction = namedtuple('FilterAction', 'user')
InteractAction = namedtuple('InteractAction', 'source_name source_type user succeed')
RemoveMassFollowerAction = namedtuple('RemoveMassFollowerAction', 'user')


@unique
class SourceType(Enum):
    BLOGGER = "blogger"
    HASHTAG = "hashtag"
    PLACE = "place"
    TARGET = "target"


@unique
class BloggerInteractionType(Enum):
    FOLLOWERS = 'followers'
    FOLLOWING = 'following'


@unique
class HashtagInteractionType(Enum):
    TOP_LIKERS = 'top-likers'
    RECENT_LIKERS = 'recent-likers'


@unique
class PlaceInteractionType(Enum):
    TOP_LIKERS = 'top-likers'
    RECENT_LIKERS = 'recent-likers'


@unique
class TargetType(Enum):
    URL = 'url'
    USERNAME = 'username'



from enum import Enum, unique

from insomniac.utils import *
from insomniac.views import TabBarView, ProfileView

SEARCH_CONTENT_DESC_REGEX = '[Ss]earch and [Ee]xplore'
SETTINGS_LIST_ID_REGEX = 'android:id/list|{0}:id/recycler_view'
SETTINGS_LIST_CLASS_NAME_REGEX = 'android.widget.ListView|androidx.recyclerview.widget.RecyclerView'


def navigate(device, tab):
    tab_name = tab.name.lower()
    tab_index = tab.value

    print("Press " + tab_name)
    if tab == Tabs.SEARCH:
        _navigate_to_search(device)
        return

    device.close_keyboard()
    tab_bar = device.find(resourceId=f'{device.app_id}:id/tab_bar', className='android.widget.LinearLayout')
    button = tab_bar.child(index=tab_index)

    # Two clicks to reset tab content
    button.click()
    button.click()


def search_for(device, username=None, hashtag=None, place=None, on_action=None):
    search_view = TabBarView(device).navigate_to_search()
    target_view = None

    if username is not None:
        target_view = search_view.navigate_to_username(username, on_action)

    if hashtag is not None:
        target_view = search_view.navigate_to_hashtag(hashtag)

    if place is not None:
        target_view = search_view.navigate_to_place(place)

    return target_view is not None


def switch_to_english(device):
    print(COLOR_OKGREEN + "Switching to English locale" + COLOR_ENDC)
    navigate(device, Tabs.PROFILE)
    ProfileView(device) \
        .navigate_to_options() \
        .navigate_to_settings() \
        .switch_to_english()


def _navigate_to_search(device):
    # Search tab is a special case, because on some accounts there is "Reels" tab instead. If so, we have to go to the
    # "Home" tab and press search in the action bar.

    device.close_keyboard()

    tab_bar = device.find(resourceId=f'{device.app_id}:id/tab_bar', className='android.widget.LinearLayout')
    search_in_tab_bar = tab_bar.child(descriptionMatches=SEARCH_CONTENT_DESC_REGEX)
    if search_in_tab_bar.exists():
        # Two clicks to reset tab content
        search_in_tab_bar.click()
        search_in_tab_bar.click()
        return

    print("Didn't find search in the tab bar...")
    navigate(device, Tabs.HOME)
    print("Press search in the action bar")
    action_bar = device.find(resourceId=f'{device.app_id}:id/action_bar', className='android.widget.LinearLayout')
    search_in_action_bar = action_bar.child(descriptionMatches=SEARCH_CONTENT_DESC_REGEX)
    if search_in_action_bar.exists():
        search_in_action_bar.click()
        return

    print(COLOR_FAIL + "Cannot find search tab neither in the tab bar, nor in the action bar. Maybe not English "
                       "language is set?" + COLOR_ENDC)
    save_crash(device)
    switch_to_english(device)
    raise LanguageChangedException()


class LanguageChangedException(Exception):
    pass


@unique
class Tabs(Enum):
    HOME = 0
    SEARCH = 1
    PLUS = 2
    LIKES = 3
    PROFILE = 4



import uuid
from datetime import timedelta
from enum import Enum, unique
from typing import Optional

from peewee import *

from insomniac.globals import task_id, execution_id
from insomniac.utils import *

DATABASE_NAME = 'insomniac.db'
DATABASE_VERSION = 1

db = SqliteDatabase(DATABASE_NAME, autoconnect=False)


class InsomniacModel(Model):
    class Meta:
        database = db


class InstagramProfile(InsomniacModel):
    name = TextField(unique=True)

    class Meta:
        db_table = 'instagram_profiles'

    def start_session(self, app_id, app_version, args, profile_status, followers_count, following_count) -> uuid.UUID:
        """
        Create InstagramProfileInfo record
        Create SessionInfo record
        Return the session id
        """
        with db.connection_context():
            profile_info = InstagramProfileInfo.create(profile=self,
                                                       status=profile_status,
                                                       followers=followers_count,
                                                       following=following_count)

            session = SessionInfo.create(app_id=app_id,
                                         app_version=app_version,
                                         args=args,
                                         profile_info=profile_info)
        return session.id

    def end_session(self, session_id):
        """
        Add end-timestamp to the SessionInfo record
        """
        with db.connection_context():
            session_info = SessionInfo.get(SessionInfo.id == session_id)
            session_info.end = datetime.now()
            session_info.save()

    def add_session(self, app_id, app_version, args, profile_status, followers_count, following_count, start, end):
        """
        For migration only!
        """
        with db.connection_context():
            profile_info = InstagramProfileInfo.create(profile=self,
                                                       status=profile_status,
                                                       followers=followers_count,
                                                       following=following_count)

            SessionInfo.create(app_id=app_id,
                               app_version=app_version,
                               args=args,
                               profile_info=profile_info,
                               start=start,
                               end=end)

    def update_profile_info(self, profile_status, followers_count, following_count):
        """
        Create a new InstagramProfileInfo record

        Can't see any usage for that right now, maybe we will use it later just in order to update profile info without
        running an insomniac-session.
        """
        with db.connection_context():
            InstagramProfileInfo.create(profile=self,
                                        status=profile_status,
                                        followers=followers_count,
                                        following=following_count)

    def log_get_profile_action(self, session_id, username, timestamp=None):
        """
        Create InsomniacAction record
        Create GetProfileAction record
        """
        with db.connection_context():
            session = SessionInfo.get(SessionInfo.id == session_id)
            action = InsomniacAction.create(actor_profile=self,
                                            type=GetProfileAction.__name__,
                                            task_id=task_id,
                                            execution_id=execution_id,
                                            session=session,
                                            timestamp=(timestamp if timestamp is not None else datetime.now()))
            GetProfileAction.create(action=action,
                                    target_user=username)

    def log_like_action(self, session_id, username, source_type, source_name, timestamp=None):
        """
        Create InsomniacAction record
        Create LikeAction record
        Create InsomniacActionSource record
        """
        with db.connection_context():
            session = SessionInfo.get(SessionInfo.id == session_id)
            action = InsomniacAction.create(actor_profile=self,
                                            type=LikeAction.__name__,
                                            task_id=task_id,
                                            execution_id=execution_id,
                                            session=session,
                                            timestamp=(timestamp if timestamp is not None else datetime.now()))
            LikeAction.create(action=action,
                              target_user=username)
            if source_type is not None and source_name is not None:
                InsomniacActionSource.create(action=action,
                                             type=source_type,
                                             name=source_name)

    def log_follow_action(self, session_id, username, source_type, source_name, timestamp=None):
        """
        Create InsomniacAction record
        Create FollowAction record
        Create InsomniacActionSource record
        """
        with db.connection_context():
            session = SessionInfo.get(SessionInfo.id == session_id)
            action = InsomniacAction.create(actor_profile=self,
                                            type=FollowAction.__name__,
                                            task_id=task_id,
                                            execution_id=execution_id,
                                            session=session,
                                            timestamp=(timestamp if timestamp is not None else datetime.now()))
            FollowAction.create(action=action,
                                target_user=username)
            if source_type is not None and source_name is not None:
                InsomniacActionSource.create(action=action,
                                             type=source_type,
                                             name=source_name)

    def log_story_watch_action(self, session_id, username, source_type, source_name, timestamp=None):
        """
        Create InsomniacAction record
        Create StoryWatchAction record
        Create InsomniacActionSource record
        """
        with db.connection_context():
            session = SessionInfo.get(SessionInfo.id == session_id)
            action = InsomniacAction.create(actor_profile=self,
                                            type=StoryWatchAction.__name__,
                                            task_id=task_id,
                                            execution_id=execution_id,
                                            session=session,
                                            timestamp=(timestamp if timestamp is not None else datetime.now()))
            StoryWatchAction.create(action=action,
                                    target_user=username)
            if source_type is not None and source_name is not None:
                InsomniacActionSource.create(action=action,
                                             type=source_type,
                                             name=source_name)

    def log_comment_action(self, session_id, username, comment, source_type, source_name, timestamp=None):
        """
        Create InsomniacAction record
        Create CommentAction record
        Create InsomniacActionSource record
        """
        with db.connection_context():
            session = SessionInfo.get(SessionInfo.id == session_id)
            action = InsomniacAction.create(actor_profile=self,
                                            type=CommentAction.__name__,
                                            task_id=task_id,
                                            execution_id=execution_id,
                                            session=session,
                                            timestamp=(timestamp if timestamp is not None else datetime.now()))
            CommentAction.create(action=action,
                                 target_user=username,
                                 comment=comment)
            if source_type is not None and source_name is not None:
                InsomniacActionSource.create(action=action,
                                             type=source_type,
                                             name=source_name)

    def log_direct_message_action(self, session_id, username, message, timestamp=None):
        """
        Create InsomniacAction record
        Create DirectMessageAction record
        """
        with db.connection_context():
            session = SessionInfo.get(SessionInfo.id == session_id)
            action = InsomniacAction.create(actor_profile=self,
                                            type=DirectMessageAction.__name__,
                                            task_id=task_id,
                                            execution_id=execution_id,
                                            session=session,
                                            timestamp=(timestamp if timestamp is not None else datetime.now()))
            DirectMessageAction.create(action=action,
                                       target_user=username,
                                       message=message)

    def log_unfollow_action(self, session_id, username, timestamp=None):
        """
        Create InsomniacAction record
        Create UnfollowAction record
        """
        with db.connection_context():
            session = SessionInfo.get(SessionInfo.id == session_id)
            action = InsomniacAction.create(actor_profile=self,
                                            type=UnfollowAction.__name__,
                                            task_id=task_id,
                                            execution_id=execution_id,
                                            session=session,
                                            timestamp=(timestamp if timestamp is not None else datetime.now()))
            UnfollowAction.create(action=action,
                                  target_user=username)

    def log_scrape_action(self, session_id, username, source_type, source_name, timestamp=None):
        """
        Create InsomniacAction record
        Create ScrapeAction record
        Create InsomniacActionSource record
        """
        with db.connection_context():
            session = SessionInfo.get(SessionInfo.id == session_id)
            action = InsomniacAction.create(actor_profile=self,
                                            type=ScrapeAction.__name__,
                                            task_id=task_id,
                                            execution_id=execution_id,
                                            session=session,
                                            timestamp=(timestamp if timestamp is not None else datetime.now()))
            ScrapeAction.create(action=action,
                                target_user=username)
            if source_type is not None and source_name is not None:
                InsomniacActionSource.create(action=action,
                                             type=(source_type if source_type is not None else None),
                                             name=source_name)

    def log_filter_action(self, session_id, username, timestamp=None):
        """
        Create InsomniacAction record
        Create FilterAction record
        """
        with db.connection_context():
            session = SessionInfo.get(SessionInfo.id == session_id)
            action = InsomniacAction.create(actor_profile=self,
                                            type=FilterAction.__name__,
                                            task_id=task_id,
                                            execution_id=execution_id,
                                            session=session,
                                            timestamp=(timestamp if timestamp is not None else datetime.now()))
            FilterAction.create(action=action,
                                target_user=username)

    def log_change_profile_info_action(self, session_id, profile_pic_url, name, description, timestamp=None):
        """
        Create InsomniacAction record
        Create ChangeProfileInfoAction record
        """
        with db.connection_context():
            session = SessionInfo.get(SessionInfo.id == session_id)
            action = InsomniacAction.create(actor_profile=self,
                                            type=ChangeProfileInfoAction.__name__,
                                            task_id=task_id,
                                            execution_id=execution_id,
                                            session=session,
                                            timestamp=(timestamp if timestamp is not None else datetime.now()))
            ChangeProfileInfoAction.create(action=action,
                                           profile_pic_url=profile_pic_url,
                                           name=name,
                                           description=description)

    def publish_scrapped_account(self, username, scrape_for_account_list):
        """
        Use this function when you are a scrapper and you have found a profile according to filters.
        Create ScrappedProfile record for each target-real-bot-profile
        """
        with db.connection_context():
            for target_actor_username in scrape_for_account_list:
                target_actor_profile, created = InstagramProfile.get_or_create(name=target_actor_username)
                if ScrappedProfile.get_or_none(target_actor_profile=target_actor_profile, name=username) is None:
                    ScrappedProfile.create(target_actor_profile=target_actor_profile, name=username)

    def update_follow_status(self, username, is_follow_me, do_i_follow_him):
        """
        Create or update FollowStatus record
        """
        with db.connection_context():
            FollowStatus.create(actor_profile=self,
                                profile=username,
                                is_follow_me=is_follow_me,
                                do_i_follow_him=do_i_follow_him)

    def _get_follow_status(self, username, hours=None):
        with db.connection_context():
            follow_statuses = FollowStatus.select() \
                .where((FollowStatus.profile == username)
                       & (FollowStatus.actor_profile == self)
                       & ((datetime.now().timestamp() - FollowStatus.updated_at <= timedelta(hours=hours).total_seconds()) if hours is not None else True)) \
                .order_by(FollowStatus.updated_at.desc()) \
                .limit(1)
            return follow_statuses[0] if len(follow_statuses) > 0 else None

    def is_follow_me(self, username, hours=None):
        follow_status = self._get_follow_status(username, hours)
        return follow_status.is_follow_me if follow_status is not None else None

    def do_i_follow(self, username, hours=None):
        follow_status = self._get_follow_status(username, hours)
        return follow_status.do_i_follow_him if follow_status is not None else None

    def used_to_follow(self, username):
        """
        Whether I used to follow this username or am following right now.
        """
        with db.connection_context():
            return len(FollowAction.select()
                       .join(InsomniacAction, join_type=JOIN.LEFT_OUTER)
                       .where((FollowAction.target_user == username)
                              & (InsomniacAction.actor_profile == self))
                       .limit(1)) > 0

    def is_interacted(self, username, hours=None) -> bool:
        with db.connection_context():
            if len(GetProfileAction.select(GetProfileAction.target_user)
                    .join(InsomniacAction, join_type=JOIN.LEFT_OUTER)
                    .where((GetProfileAction.target_user == username)
                           & (InsomniacAction.actor_profile == self)
                           & ((datetime.now().timestamp() - InsomniacAction.timestamp <= timedelta(hours=hours).total_seconds()) if hours is not None else True))
                    .limit(1)) > 0:
                return True

            if len(LikeAction.select(LikeAction.target_user)
                    .join(InsomniacAction, join_type=JOIN.LEFT_OUTER)
                    .where((LikeAction.target_user == username)
                           & (InsomniacAction.actor_profile == self)
                           & ((datetime.now().timestamp() - InsomniacAction.timestamp <= timedelta(hours=hours).total_seconds()) if hours is not None else True))
                    .limit(1)) > 0:
                return True

            if len(CommentAction.select(CommentAction.target_user)
                    .join(InsomniacAction, join_type=JOIN.LEFT_OUTER)
                    .where((CommentAction.target_user == username)
                           & (InsomniacAction.actor_profile == self)
                           & ((datetime.now().timestamp() - InsomniacAction.timestamp <= timedelta(hours=hours).total_seconds()) if hours is not None else True))
                    .limit(1)) > 0:
                return True
        return False

    def is_filtered(self, username, hours=None) -> bool:
        with db.connection_context():
            if len(FilterAction.select(FilterAction.target_user)
                    .join(InsomniacAction, join_type=JOIN.LEFT_OUTER)
                    .where((FilterAction.target_user == username)
                           & (InsomniacAction.actor_profile == self)
                           & ((datetime.now().timestamp() - InsomniacAction.timestamp <= timedelta(hours=hours).total_seconds()) if hours is not None else True))
                    .limit(1)) > 0:
                return True
        return False

    def is_scrapped(self, username, scrape_for_account_list) -> bool:
        """
        Return True if "username" is scrapped for ALL accounts from "scrape_for_account_list", False otherwise.
        """
        with db.connection_context():
            for target_actor_profile in scrape_for_account_list:
                if len(ScrappedProfile.select()
                       .where(ScrappedProfile.name == username)
                       .join(InstagramProfile, join_type=JOIN.LEFT_OUTER)
                       .where(InstagramProfile.name == target_actor_profile)
                       .limit(1)) == 0:
                    return False
            return True

    def get_scrapped_profile_for_interaction(self) -> Optional[str]:
        """
        Use this function when you are interacting with targets, and you are looking for the next scrapped target
        """
        with db.connection_context():
            scrapped_profiles = ScrappedProfile.select(ScrappedProfile.name) \
                .where(ScrappedProfile.target_actor_profile == self) \
                .join(GetProfileAction, join_type=JOIN.LEFT_OUTER, on=(ScrappedProfile.name == GetProfileAction.target_user)) \
                .switch(ScrappedProfile) \
                .join(LikeAction, join_type=JOIN.LEFT_OUTER, on=(ScrappedProfile.name == LikeAction.target_user)) \
                .switch(ScrappedProfile) \
                .join(FollowAction, join_type=JOIN.LEFT_OUTER, on=(ScrappedProfile.name == FollowAction.target_user)) \
                .switch(ScrappedProfile) \
                .join(StoryWatchAction, join_type=JOIN.LEFT_OUTER, on=(ScrappedProfile.name == StoryWatchAction.target_user)) \
                .switch(ScrappedProfile) \
                .join(CommentAction, join_type=JOIN.LEFT_OUTER, on=(ScrappedProfile.name == CommentAction.target_user)) \
                .where(
                    GetProfileAction.target_user.is_null()
                    & LikeAction.target_user.is_null()
                    & FollowAction.target_user.is_null()
                    & StoryWatchAction.target_user.is_null()
                    & CommentAction.target_user.is_null()
                ) \
                .order_by(ScrappedProfile.timestamp.desc()) \
                .limit(1)

            return scrapped_profiles[0].name if len(scrapped_profiles) > 0 else None

    def count_scrapped_profiles_for_interaction(self):
        with db.connection_context():
            return len(ScrappedProfile.select(ScrappedProfile.name)
                       .where(ScrappedProfile.target_actor_profile == self)
                       .join(GetProfileAction, join_type=JOIN.LEFT_OUTER, on=(ScrappedProfile.name == GetProfileAction.target_user))
                       .switch(ScrappedProfile)
                       .join(LikeAction, join_type=JOIN.LEFT_OUTER, on=(ScrappedProfile.name == LikeAction.target_user))
                       .switch(ScrappedProfile)
                       .join(FollowAction, join_type=JOIN.LEFT_OUTER, on=(ScrappedProfile.name == FollowAction.target_user))
                       .switch(ScrappedProfile)
                       .join(StoryWatchAction, join_type=JOIN.LEFT_OUTER, on=(ScrappedProfile.name == StoryWatchAction.target_user))
                       .switch(ScrappedProfile)
                       .join(CommentAction, join_type=JOIN.LEFT_OUTER, on=(ScrappedProfile.name == CommentAction.target_user))
                       .where(GetProfileAction.target_user.is_null()
                              & LikeAction.target_user.is_null()
                              & FollowAction.target_user.is_null()
                              & StoryWatchAction.target_user.is_null()
                              & CommentAction.target_user.is_null()))


class InstagramProfileInfo(InsomniacModel):
    profile = ForeignKeyField(InstagramProfile, backref='profile_info_records')
    status = TextField()
    followers = BigIntegerField()
    following = BigIntegerField()
    timestamp = TimestampField(default=datetime.now)

    class Meta:
        db_table = 'instagram_profiles_info'


class SessionInfo(InsomniacModel):
    app_id = TextField(null=True)
    app_version = TextField()
    start = TimestampField(default=datetime.now)
    end = TimestampField(null=True)
    args = TextField()
    profile_info = ForeignKeyField(InstagramProfileInfo, backref='session_info')

    class Meta:
        db_table = 'sessions_info'


class InsomniacAction(InsomniacModel):
    actor_profile = ForeignKeyField(InstagramProfile, backref='actions')
    type = TextField()
    timestamp = TimestampField(default=datetime.now)
    task_id = UUIDField()
    execution_id = UUIDField()
    session = ForeignKeyField(SessionInfo, backref='session_actions')

    class Meta:
        db_table = 'actions_log'


class InsomniacActionSource(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='action_source')
    type = TextField()
    name = TextField()

    class Meta:
        db_table = 'actions_sources'


class GetProfileAction(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='get_profile_action_info')
    target_user = TextField()

    class Meta:
        db_table = 'get_profile_actions'


class LikeAction(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='like_action_info')
    target_user = TextField()

    class Meta:
        db_table = 'like_actions'


class FollowAction(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='follow_action_info')
    target_user = TextField()

    class Meta:
        db_table = 'follow_actions'


class StoryWatchAction(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='story_watch_action_info')
    target_user = TextField()

    class Meta:
        db_table = 'story_watch_actions'


class CommentAction(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='comment_action_info')
    target_user = TextField()
    comment = TextField()

    class Meta:
        db_table = 'comment_actions'


class DirectMessageAction(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='direct_message_action_info')
    target_user = TextField()
    message = TextField()

    class Meta:
        db_table = 'direct_message_actions'


class UnfollowAction(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='unfollow_action_info')
    target_user = TextField()

    class Meta:
        db_table = 'unfollow_actions'


class ScrapeAction(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='scrape_action_info')
    target_user = TextField()

    class Meta:
        db_table = 'scrape_actions'


class FilterAction(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='filter_action_info')
    target_user = TextField()

    class Meta:
        db_table = 'filter_actions'


class ChangeProfileInfoAction(InsomniacModel):
    action = ForeignKeyField(InsomniacAction, backref='change_profile_info_action_info')
    profile_pic_url = TextField()
    name = TextField()
    description = TextField()

    class Meta:
        db_table = 'change_profile_info_actions'


class ScrappedProfile(InsomniacModel):
    # "target actor" is a profile which will actually do the interaction with the target
    target_actor_profile = ForeignKeyField(InstagramProfile, backref='scrapped_profiles')
    name = TextField()
    timestamp = TimestampField(default=datetime.now)

    class Meta:
        db_table = 'scrapped_profiles'


class FollowStatus(InsomniacModel):
    # "actor" is a profile that actually did the interaction and now updates he database
    actor_profile = ForeignKeyField(InstagramProfile)
    profile = TextField()
    is_follow_me = BooleanField(null=True, default=None)
    do_i_follow_him = BooleanField(null=True, default=None)
    updated_at = TimestampField(default=datetime.now)

    class Meta:
        db_table = 'follow_status'


class SchemaVersion(InsomniacModel):
    version = SmallIntegerField(default=DATABASE_VERSION)
    updated_at = TimestampField(default=datetime.now)

    class Meta:
        db_table = 'schema_version'


@unique
class ProfileStatus(Enum):
    VALID = "valid"
    # TODO: request list of possible statuses from Jey


MODELS = [
    SchemaVersion,
    InstagramProfile,
    InstagramProfileInfo,
    SessionInfo,
    InsomniacAction,
    InsomniacActionSource,
    GetProfileAction,
    LikeAction,
    FollowAction,
    StoryWatchAction,
    CommentAction,
    DirectMessageAction,
    UnfollowAction,
    ScrapeAction,
    FilterAction,
    ChangeProfileInfoAction,
    ScrappedProfile,
    FollowStatus
]


def init() -> bool:
    """
    Initialize database and return whether it was just created.
    """
    with db.connection_context():
        # Migration logic between schema versions can be added here later
        if len(db.get_tables()) == 0:
            print(f"Creating tables in {DATABASE_NAME}...")
            db.create_tables(MODELS)
            SchemaVersion.create()
            return True
    return False


def get_ig_profile_by_profile_name(profile_name) -> InstagramProfile:
    with db.connection_context():
        profile, is_created = InstagramProfile.get_or_create(name=profile_name)
    return profile



import base64

import insomniac
from insomniac.utils import *

# Typewriter uses Android application (apk file) built from this repo: https://github.com/alexal1/InsomniacAutomator
# It provides IME (Input Method Editor) that replaces virtual keyboard with it's own one, which listens to specific
# broadcast messages and simulates key presses.
ADB_KEYBOARD_IME = "com.alexal1.adbkeyboard/.AdbIME"
ADB_KEYBOARD_APK = "ADBKeyboard.apk"
DELAY_MEAN = 200
DELAY_DEVIATION = 100
IME_MESSAGE_B64 = "ADB_INPUT_B64"
IME_CLEAR_TEXT = "ADB_CLEAR_TEXT"
EXTRA_MESSAGE = "msg"
EXTRA_DELAY_MEAN = "delay_mean"
EXTRA_DELAY_DEVIATION = "delay_deviation"


class Typewriter:
    is_adb_keyboard_set = False
    device_id = None

    def __init__(self, device_id):
        self.is_adb_keyboard_set = False
        self.device_id = device_id

    def set_adb_keyboard(self):
        if not self._is_adb_ime_existing():
            print("Installing ADB Keyboard to enable typewriting...")
            apk_path = os.path.join(os.path.dirname(os.path.abspath(insomniac.__file__)), "apk", ADB_KEYBOARD_APK)
            os.popen("adb" + ("" if self.device_id is None else " -s " + self.device_id)
                     + f" install {apk_path}").close()
        self.is_adb_keyboard_set = self._set_adb_ime()
        if not self.is_adb_keyboard_set:
            print(COLOR_FAIL + "Cannot setup ADB Keyboard. Don't worry! Fallback to text copy-pasting will be used."
                  + COLOR_ENDC)

    def write(self, view, text) -> bool:
        if not self.is_adb_keyboard_set:
            return False
        view.click()
        if not self.clear():
            return False
        text_b64 = base64.b64encode(text.encode('utf-8')).decode('utf-8')
        extras = {
            EXTRA_MESSAGE: text_b64,
            EXTRA_DELAY_MEAN: DELAY_MEAN,
            EXTRA_DELAY_DEVIATION: DELAY_DEVIATION
        }
        self._send_broadcast(IME_MESSAGE_B64, extras)
        sleep_millis = DELAY_MEAN * len(text) + DELAY_DEVIATION
        sleep_secs = sleep_millis / 1000.0
        print_debug(f"Wait until text is typed: {sleep_secs} seconds")
        sleep(sleep_secs)
        return True

    def clear(self) -> bool:
        if not self.is_adb_keyboard_set:
            return False
        self._send_broadcast(IME_CLEAR_TEXT)
        return True

    def _set_adb_ime(self):
        attempts_count = 0
        while not self._is_adb_ime_existing():
            attempts_count += 1
            if attempts_count == 5:
                return False
            sleep(2)
        stream = os.popen("adb" + ("" if self.device_id is None else " -s " + self.device_id)
                          + f" shell ime set {ADB_KEYBOARD_IME}")
        output = stream.read()
        succeed = "selected" in output
        stream.close()
        return succeed

    def _is_adb_ime_existing(self):
        stream = os.popen("adb" + ("" if self.device_id is None else " -s " + self.device_id) + " shell ime list -a")
        output = stream.read()
        result = ADB_KEYBOARD_IME in output
        stream.close()
        return result

    def _send_broadcast(self, action, extras=None):
        command = "adb" + ("" if self.device_id is None else " -s " + self.device_id) \
                  + f" shell am broadcast -a {action}"
        if extras is not None:
            for key, value in extras.items():
                if isinstance(value, str):
                    command += f" --es {key} '{value}'"
                elif isinstance(value, int):
                    command += f" --ei {key} {value}"
                else:
                    print_debug(COLOR_FAIL + f"Unexpected broadcast extra: {value}" + COLOR_ENDC)
                    continue
        os.popen(command).close()



# This file is deprecated. We use peewee models (see db_models.py) since v3.7.1

import sqlite3
from enum import Enum, unique

from insomniac.actions_providers import Provider
from insomniac.utils import *

DB_NAME = "interaction_data.db"
DB_VERSIONS = {'3.5.0': 1,
               '3.5.18': 2}

SQL_SELECT_FROM_METADATA = "SELECT * from metadata"
SQL_SELECT_MAX_VERSION_FROM_METADATA = "SELECT MAX(version) from metadata"
SQL_SELECT_FROM_FOLLOW_STATUS_BY_USERNAME = "SELECT * from users_follow_status WHERE username = :username"
SQL_SELECT_FROM_INTERACTED_USERS_BY_USERNAME = "SELECT * from interacted_users WHERE username = :username"
SQL_SELECT_TARGETS_FROM_INTERACTED_USERS = "SELECT * from interacted_users WHERE " \
                                           "provider = 'TARGETS_LIST' OR provider = 'SCRAPING'"
SQL_COUNT_LOADED_TARGETS_FROM_INTERACTED_USERS_BY_SCRAPE = "SELECT COUNT(*) from interacted_users WHERE provider = 'SCRAPING'"
SQL_COUNT_LOADED_TARGETS_FROM_INTERACTED_USERS_BY_TARGETS_LIST = "SELECT COUNT(*) from interacted_users WHERE provider = 'TARGETS_LIST'"
SQL_COUNT_LOADED_NOT_INTERACTED_TARGETS_FROM_INTERACTED_USERS_BY_SCRAPE = "SELECT COUNT(*) from interacted_users WHERE " \
                                                                          "provider = 'SCRAPING' AND interactions_count = 0"
SQL_COUNT_LOADED_NOT_INTERACTED_TARGETS_FROM_INTERACTED_USERS_BY_TARGETS_LIST = "SELECT COUNT(*) from interacted_users WHERE " \
                                                                                "provider = 'TARGETS_LIST' AND interactions_count = 0"
SQL_SELECT_FROM_FILTERED_USERS_BY_USERNAME = "SELECT * from filtered_users WHERE username = :username"
SQL_SELECT_FROM_SCRAPED_USERS_BY_USERNAME = "SELECT * from scraped_users WHERE username = :username"
SQL_SELECT_ALL_SESSIONS = "SELECT * from sessions INNER JOIN profiles ON sessions.profile_id = profiles.id"
SQL_SELECT_ALL_INTERACTED_USERS = "SELECT * from interacted_users"
SQL_SELECT_ALL_FILTERED_USERS = "SELECT * from filtered_users"
SQL_SELECT_ALL_SCRAPED_USERS = "SELECT * from scraped_users"
SQL_SELECT_PROFILE_BY_ID = "SELECT * from profiles WHERE id = :profile_id"

SQL_INSERT_DEFAULT_INTO_METADATA = "INSERT INTO metadata DEFAULT VALUES"
SQL_INSERT_INTO_METADATA = "INSERT INTO metadata (version) VALUES (?)"
SQL_INSERT_INTO_INTERACTED_USERS = "INSERT INTO interacted_users (username, last_interaction, " \
                                   "source, interaction_type, provider) VALUES (?, ?, ?, ?, ?)"
SQL_INSERT_INTO_FILTERED_USERS = "INSERT INTO filtered_users (username, filtered_at) VALUES (?, ?)"
SQL_INSERT_INTO_SCRAPED_USERS = "INSERT INTO scraped_users (username, last_interaction) VALUES (?, ?)"
SQL_INSERT_INTO_PROFILES = "INSERT INTO profiles (followers, following) VALUES (?, ?)"
SQL_INSERT_INTO_SESSIONS = """
    INSERT INTO sessions (
        app_version,
        total_interactions,
        successful_interactions,
        total_followed,
        total_likes,
        total_unfollowed,
        total_stories_watched,
        total_get_profile,
        total_scraped,
        removed_mass_followers,
        start_time,
        finish_time,
        args,
        profile_id
    ) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"""
SQL_INSERT_INTO_FOLLOW_STATUS = "INSERT INTO users_follow_status (username, is_follow_me, " \
                                   "do_i_follow_him, updated_at) VALUES (?, ?, ?, ?)"

SQL_DELETE_FROM_TARGETS_BY_USERNAME = "DELETE from targets WHERE username = :username"

SQL_UPDATE_INTERACTED_USER = "UPDATE interacted_users set following_status = ?, last_interaction = ?, " \
                             "interactions_count = ?, source = ?, interaction_type = ?, provider = ? where username = ?"
SQL_UPDATE_INTERACTED_USER_INTERACTIONS_COUNT = "UPDATE interacted_users set interactions_count = ? where username = ?"
SQL_UPDATE_FILTERED_USER = "UPDATE filtered_users set filtered_at = ? where username = ?"
SQL_UPDATE_SCRAPED_USER = "UPDATE scraped_users set scraping_status = ?, last_interaction = ? where username = ?"
SQL_UPDATE_FOLLOW_STATUS = "UPDATE users_follow_status set is_follow_me = ?, do_i_follow_him = ?, " \
                           "updated_at = ? where username = ?"

SQL_CREATE_METADATA_TABLE = f"""
    CREATE TABLE IF NOT EXISTS `metadata` (
        `version` INTEGER PRIMARY KEY DEFAULT {max(DB_VERSIONS.values())});"""

SQL_CREATE_INTERACTED_USERS_TABLE = """
    CREATE TABLE IF NOT EXISTS `interacted_users` (
        `username` TEXT PRIMARY KEY,
        `following_status` TEXT CHECK(`following_status` IN ('NONE', 'FOLLOWED', 'UNFOLLOWED')) NOT NULL DEFAULT 'NONE',
        `last_interaction` DATETIME NOT NULL,
        `interactions_count` INTEGER NOT NULL DEFAULT 0,
        `source` TEXT,
        `interaction_type` TEXT,
        `provider` TEXT CHECK(`provider` IN ('UNKNOWN', 'INTERACTION', 'SCRAPING', 'TARGETS_LIST')));"""

SQL_CREATE_SCRAPED_USERS_TABLE = """
    CREATE TABLE IF NOT EXISTS `scraped_users` (
        `username` TEXT PRIMARY KEY,
        `scraping_status` TEXT CHECK(`scraping_status` IN ('SCRAPED','NOT_SCRAPED')) NOT NULL DEFAULT 'NOT_SCRAPED',
        `last_interaction` DATETIME NOT NULL);"""

SQL_CREATE_FILTERED_USERS_TABLE = """
    CREATE TABLE IF NOT EXISTS `filtered_users` (
        `username` TEXT PRIMARY KEY,
        `filtered_at` DATETIME NOT NULL);"""

SQL_CREATE_PROFILE_TABLE = """
    CREATE TABLE IF NOT EXISTS `profiles` (
        `id` INTEGER PRIMARY KEY AUTOINCREMENT,
        `followers` INTEGER,
        `following` INTEGER);"""

SQL_CREATE_SESSIONS_TABLE = """
    CREATE TABLE IF NOT EXISTS `sessions` (
        `id` INTEGER PRIMARY KEY AUTOINCREMENT,
        `app_version` TEXT NOT NULL,
        `total_interactions` INTEGER NOT NULL,
        `successful_interactions` INTEGER NOT NULL,
        `total_followed` INTEGER NOT NULL,
        `total_likes` INTEGER NOT NULL,
        `total_unfollowed` INTEGER NOT NULL,
        `total_stories_watched` INTEGER NOT NULL,
        `total_get_profile` INTEGER NOT NULL,
        `total_scraped` INTEGER NOT NULL,
        `removed_mass_followers` TEXT NOT NULL,
        `start_time` DATETIME NOT NULL,
        `finish_time` DATETIME,
        `args` TEXT NOT NULL,
        `profile_id` INTEGER REFERENCES `profiles` (id));"""

SQL_CREATE_USERS_FOLLOW_STATUS_TABLE = """
    CREATE TABLE IF NOT EXISTS `users_follow_status` (
        `username` TEXT PRIMARY KEY,
        `is_follow_me` TEXT CHECK(`is_follow_me` IN ('UNKNOWN', 'TRUE', 'FALSE')) NOT NULL DEFAULT 'UNKNOWN',
        `do_i_follow_him` TEXT CHECK(`do_i_follow_him` IN ('UNKNOWN', 'TRUE', 'FALSE')) NOT NULL DEFAULT 'UNKNOWN',
        `updated_at` DATETIME NOT NULL);"""


def get_database(username):
    address = os.path.join(username, DB_NAME)
    if not check_database_exists(username):
        create_database(address)
    else:
        migrate_database_if_needed(address)
    return address


def check_database_exists(username, create_db_dir=True):
    address = os.path.join(username, DB_NAME)
    verify_database_directories(address, create_db_dir)
    return os.path.isfile(address)


def migrate_database_if_needed(address):
    connection = None
    try:
        connection = sqlite3.connect(address)
        with connection:
            connection.row_factory = sqlite3.Row
            cursor = connection.cursor()
            _run_migrations(cursor)
            connection.commit()
    except DatabaseMigrationFailedException as e:
        raise e
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot create/open database at {address}: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()


def create_database(address):
    connection = None
    try:
        connection = sqlite3.connect(address)
        with connection:
            connection.row_factory = sqlite3.Row
            cursor = connection.cursor()
            create_tables(
                cursor,
                [
                    "metadata",
                    "interacted_users",
                    "scraped_users",
                    "filtered_users",
                    "profiles",
                    "sessions",
                    "targets",
                    "users_follow_status"
                ]
            )
            cursor.execute(SQL_INSERT_DEFAULT_INTO_METADATA)
            connection.commit()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot create/open database at {address}: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()


def create_tables(cursor, tables):
    if "metadata" in tables:
        cursor.execute(SQL_CREATE_METADATA_TABLE)

    if "interacted_users" in tables:
        cursor.execute(SQL_CREATE_INTERACTED_USERS_TABLE)

    if "scraped_users" in tables:
        cursor.execute(SQL_CREATE_SCRAPED_USERS_TABLE)

    if "filtered_users" in tables:
        cursor.execute(SQL_CREATE_FILTERED_USERS_TABLE)

    if "profiles" in tables:
        cursor.execute(SQL_CREATE_PROFILE_TABLE)

    if "sessions" in tables:
        cursor.execute(SQL_CREATE_SESSIONS_TABLE)

    if "users_follow_status" in tables:
        cursor.execute(SQL_CREATE_USERS_FOLLOW_STATUS_TABLE)


def verify_database_directories(address, create_db_dir=True):
    db_dir = os.path.dirname(address)
    if not os.path.exists(db_dir) and create_db_dir:
        os.makedirs(db_dir)


def get_interacted_user(address, username):
    connection = None
    interacted_user = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        interacted_user = _select_interacted_user_by_username(cursor, username)
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot get interacted user {username}: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()

    return dict(interacted_user) if interacted_user is not None else None


def get_all_interacted_users(address):
    connection = None
    interacted_users = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = dict_factory
        cursor = connection.cursor()
        cursor.execute(SQL_SELECT_ALL_INTERACTED_USERS)
        interacted_users = cursor.fetchall()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot get all interacted users: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()

    return list(interacted_users) if interacted_users is not None else ()


def update_interacted_users(address,
                            usernames,
                            last_interactions,
                            following_statuses,
                            sources,
                            interaction_types,
                            providers):
    connection = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        for i, username in enumerate(usernames):
            last_interaction = last_interactions[i]
            following_status = following_statuses[i]
            source = sources[i]
            interaction_type = interaction_types[i]
            provider = providers[i]

            interacted_user = _select_interacted_user_by_username(cursor, username)
            if interacted_user is None:
                cursor.execute(SQL_INSERT_INTO_INTERACTED_USERS, (username, last_interaction, source, interaction_type, provider.name))
            cursor.execute(
                SQL_UPDATE_INTERACTED_USER,
                (
                    following_status.name,
                    last_interaction,
                    interacted_user["interactions_count"] + 1 if interacted_user is not None else 1,
                    source if source is not None else (interacted_user["source"] if interacted_user is not None else None),
                    interaction_type if interaction_type is not None else (interacted_user["interaction_type"] if interacted_user is not None else None),
                    provider.name if provider is not None else (interacted_user["provider"] if interacted_user is not None else Provider.UNKNOWN),
                    username
                )
            )
        connection.commit()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot update interacted users: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()


def get_user_follow_status(address, username):
    connection = None
    follow_status = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        follow_status = _select_follow_status_by_username(cursor, username)
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot get user's follow status of {username}: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()

    return dict(follow_status) if follow_status is not None else None


def update_user_follow_status(address,
                              username,
                              is_follow_me,
                              do_i_follow_him,
                              updated_at):
    connection = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        interacted_user = _select_follow_status_by_username(cursor, username)
        if interacted_user is None:
            cursor.execute(SQL_INSERT_INTO_FOLLOW_STATUS, (username, 'UNKNOWN', 'UNKNOWN', updated_at))
        cursor.execute(
            SQL_UPDATE_FOLLOW_STATUS,
            (
                'TRUE' if is_follow_me else 'FALSE',
                'TRUE' if do_i_follow_him else 'FALSE',
                updated_at,
                username
            )
        )
        connection.commit()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot update user's follow-status: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()


def get_filtered_user(address, username):
    connection = None
    filtered_user = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        filtered_user = _select_filtered_user_by_username(cursor, username)
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot get filtered user {username}: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()

    return dict(filtered_user) if filtered_user is not None else None


def get_all_filtered_users(address):
    connection = None
    filtered_users = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = dict_factory
        cursor = connection.cursor()
        cursor.execute(SQL_SELECT_ALL_FILTERED_USERS)
        filtered_users = cursor.fetchall()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot get all filtered users: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()

    return list(filtered_users) if filtered_users is not None else ()


def update_filtered_users(address, usernames, filtered_at_list):
    connection = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        for i, username in enumerate(usernames):
            filtered_at = filtered_at_list[i]

            filtered_user = _select_filtered_user_by_username(cursor, username)
            if filtered_user is None:
                cursor.execute(SQL_INSERT_INTO_FILTERED_USERS, (username, filtered_at))
            cursor.execute(SQL_UPDATE_FILTERED_USER, (filtered_at, username))
        connection.commit()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot update filtered users: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()


def get_scraped_user(address, username):
    connection = None
    scraped_user = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        scraped_user = _select_scraped_user_by_username(cursor, username)
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot get scraped user {username}: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()

    return dict(scraped_user) if scraped_user is not None else None


def get_all_scraped_users(address):
    connection = None
    scraped_users = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = dict_factory
        cursor = connection.cursor()
        cursor.execute(SQL_SELECT_ALL_FILTERED_USERS)
        scraped_users = cursor.fetchall()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot get all filtered users: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()

    return list(scraped_users) if scraped_users is not None else ()


def update_scraped_users(address, usernames, last_interactions, scraping_statuses):
    connection = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        for i, username in enumerate(usernames):
            last_interaction = last_interactions[i]
            scraping_status = scraping_statuses[i]

            scraped_user = _select_scraped_user_by_username(cursor, username)
            if scraped_user is None:
                cursor.execute(SQL_INSERT_INTO_SCRAPED_USERS, (username, last_interaction))
            cursor.execute(SQL_UPDATE_SCRAPED_USER, (scraping_status.name, last_interaction, username))
        connection.commit()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot update scraped users: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()


def add_targets(address, usernames, provider, source=None, interaction_type=None):
    connection = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        for username in usernames:
            user = _select_interacted_user_by_username(cursor, username)
            if user is None:
                cursor.execute(
                    SQL_INSERT_INTO_INTERACTED_USERS,
                    (
                        username,
                        datetime.now(),
                        source,
                        interaction_type,
                        provider.name
                    )
                )
        connection.commit()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot add targets: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()


def get_target(address, user_false_validators):
    """
    Takes a user from interacted_user table which satisfies the following conditions:
    1. Has zero interactions_count
    2. Provider is either TARGETS_LIST or SCRAPING
    Then increments interactions_count.

    :return: username or None if no such user found
    """
    connection = None
    target = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        cursor.execute(SQL_SELECT_TARGETS_FROM_INTERACTED_USERS)
        target = cursor.fetchone()
        while target is not None:
            is_user_validated = True
            for validator in user_false_validators:
                if validator(target["username"]):
                    is_user_validated = False
                    break

            if is_user_validated:
                # Wasn't filtered before / blacklisted so returning that one
                break

            target = cursor.fetchone()

    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot pop target: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()

    target_username = None
    provider = None
    if target is not None:
        target_username = target["username"]
        provider = Provider.SCRAPING if target["provider"] == "SCRAPING" else Provider.TARGETS_LIST
    return target_username, provider


def count_targets(address):
    connection = None
    not_interacted_targets = {"scraped": None, "targets": None}
    total_loaded_targets = {"scraped": None, "targets": None}
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        cursor.execute(SQL_COUNT_LOADED_TARGETS_FROM_INTERACTED_USERS_BY_SCRAPE)
        total_loaded_targets["scraped"] = cursor.fetchone()[0]
        cursor = connection.cursor()
        cursor.execute(SQL_COUNT_LOADED_TARGETS_FROM_INTERACTED_USERS_BY_TARGETS_LIST)
        total_loaded_targets["targets"] = cursor.fetchone()[0]
        cursor = connection.cursor()
        cursor.execute(SQL_COUNT_LOADED_NOT_INTERACTED_TARGETS_FROM_INTERACTED_USERS_BY_SCRAPE)
        not_interacted_targets["scraped"] = cursor.fetchone()[0]
        cursor = connection.cursor()
        cursor.execute(SQL_COUNT_LOADED_NOT_INTERACTED_TARGETS_FROM_INTERACTED_USERS_BY_TARGETS_LIST)
        not_interacted_targets["targets"] = cursor.fetchone()[0]
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot count targets target: {e}" + COLOR_ENDC)
        total_loaded_targets = None
        not_interacted_targets = None
    finally:
        if connection:
            # Close the opened connection
            connection.close()

    return total_loaded_targets, not_interacted_targets


def add_sessions(address, session_states):
    connection = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = sqlite3.Row
        cursor = connection.cursor()
        for session_state in session_states:
            _add_session(cursor, session_state)
        connection.commit()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot add sessions: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()


def get_all_sessions(address):
    connection = None
    sessions = None
    try:
        connection = sqlite3.connect(address)
        connection.row_factory = dict_factory
        cursor = connection.cursor()
        cursor.execute(SQL_SELECT_ALL_SESSIONS)
        sessions = cursor.fetchall()
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Cannot get all sessions: {e}" + COLOR_ENDC)
    finally:
        if connection:
            # Close the opened connection
            connection.close()

    return list(sessions) if sessions is not None else ()


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


def _run_migrations(cursor):
    current_version = _get_database_version(cursor)
    latest_version = max(DB_VERSIONS.values())

    if current_version is None:
        return

    if current_version > latest_version:
        raise Exception(f"[Database] Current DB version (v{current_version}) is newer from your Insomniac version and "
                        f"not supported. Please update Insomniac.")

    if current_version < latest_version:
        print(f"[Database] Going to migrate database to a newer version...")
        while current_version < latest_version:
            _migrate(current_version, cursor)
            current_version = _get_database_version(cursor)


def _get_database_version(cursor):
    cursor.execute(SQL_SELECT_MAX_VERSION_FROM_METADATA)
    metadata_row = cursor.fetchone()
    if metadata_row is None:
        print(COLOR_FAIL + f"[Database] Couldn't find database-version on metadata. "
                           f"Using database as is (without trying to migrate)." + COLOR_ENDC)
        return None
    return dict(metadata_row)["MAX(version)"]


def _select_interacted_user_by_username(cursor, username):
    cursor.execute(SQL_SELECT_FROM_INTERACTED_USERS_BY_USERNAME, {"username": username})
    interacted_user = cursor.fetchone()
    return dict(interacted_user) if interacted_user is not None else None


def _select_follow_status_by_username(cursor, username):
    cursor.execute(SQL_SELECT_FROM_FOLLOW_STATUS_BY_USERNAME, {"username": username})
    user_follow_status = cursor.fetchone()
    return dict(user_follow_status) if user_follow_status is not None else None


def _select_filtered_user_by_username(cursor, username):
    cursor.execute(SQL_SELECT_FROM_FILTERED_USERS_BY_USERNAME, {"username": username})
    filtered_user = cursor.fetchone()
    return filtered_user


def _select_scraped_user_by_username(cursor, username):
    cursor.execute(SQL_SELECT_FROM_SCRAPED_USERS_BY_USERNAME, {"username": username})
    scraped_user = cursor.fetchone()
    return scraped_user


def _add_session(cursor, session_state):
    cursor.execute(SQL_INSERT_INTO_PROFILES, (session_state.my_followers_count, session_state.my_following_count))
    profile_id = cursor.lastrowid
    cursor.execute(
        SQL_INSERT_INTO_SESSIONS,
        (
            session_state.app_version,
            sum(session_state.totalInteractions.values()),
            sum(session_state.successfulInteractions.values()),
            sum(session_state.totalFollowed.values()),
            session_state.totalLikes,
            session_state.totalUnfollowed,
            session_state.totalStoriesWatched,
            session_state.totalGetProfile,
            sum(session_state.totalScraped.values()),
            str(session_state.removedMassFollowers),
            session_state.startTime,
            session_state.finishTime,
            str(session_state.args),
            profile_id
        )
    )


def _migrate_db_from_version_1_to_2(cursor):
    """
    Changes added on DB version 2:
      * Added users_follow_status database

    """

    create_tables(
        cursor,
        [
            "users_follow_status"
        ]
    )


def _migrate(curr_version, cursor):
    print(f"[Database] Going to run database migration from version {curr_version} to {curr_version+1}")

    migration_method = database_migrations[f"{curr_version}->{curr_version + 1}"]
    try:
        migration_method(cursor)
    except Exception as e:
        print(COLOR_FAIL + f"[Database] Got an error while migrating database from version "
                           f"{curr_version} to {curr_version + 1}, Error: {str(e)}" + COLOR_ENDC)
        raise DatabaseMigrationFailedException()

    print(f"[Database] database migration from version {curr_version} to {curr_version+1} has been done successfully")
    print(f"[Database] Updating database version to be {curr_version+1}")
    cursor.execute(SQL_INSERT_INTO_METADATA, (curr_version+1,))


database_migrations = {
    "1->2": _migrate_db_from_version_1_to_2
}


class DatabaseMigrationFailedException(Exception):
    pass


@unique
class ScrappingStatus(Enum):
    SCRAPED = 0
    NOT_SCRAPED = 1
